from typing import Any

from hatchet_sdk.runnables.workflow import BaseWorkflow
from hatchet_sdk.worker.slot_types import SlotType


def normalize_slot_config(
    slot_config: dict[SlotType | str, int],
) -> dict[str, int]:
    return {
        (key.value if isinstance(key, SlotType) else key): value
        for key, value in slot_config.items()
    }


def has_slot_config(
    slot_config: dict[SlotType | str, int], slot_type: SlotType
) -> bool:
    return slot_type in slot_config or slot_type.value in slot_config


def ensure_slot_config(
    slot_config: dict[SlotType | str, int], slot_type: SlotType, default_value: int
) -> dict[SlotType | str, int]:
    if has_slot_config(slot_config, slot_type):
        return slot_config
    return {**slot_config, slot_type: default_value}


def required_slot_types_from_workflows(
    workflows: list[BaseWorkflow[Any]] | None,
) -> set[str]:
    required: set[str] = set()
    if not workflows:
        return required

    for workflow in workflows:
        for task in workflow.tasks:
            if task.is_durable:
                required.add(SlotType.DURABLE.value)
            for key in task.slot_requests:
                required.add(key.value if isinstance(key, SlotType) else key)

    return required


def resolve_worker_slot_config(
    slot_config: dict[SlotType | str, int] | None,
    slots: int | None,
    durable_slots: int | None,
    workflows: list[BaseWorkflow[Any]] | None,
) -> dict[SlotType | str, int]:
    resolved_config: dict[SlotType | str, int]

    if slot_config is not None:
        resolved_config = slot_config
    else:
        legacy_config: dict[SlotType | str, int] = {
            key: value
            for key, value in (
                (SlotType.DEFAULT, slots),
                (SlotType.DURABLE, durable_slots),
            )
            if value is not None
        }
        resolved_config = legacy_config if legacy_config else {}

    required_slot_types = required_slot_types_from_workflows(workflows)

    # Apply defaults for well-known slot types
    if SlotType.DEFAULT.value in required_slot_types:
        resolved_config = ensure_slot_config(resolved_config, SlotType.DEFAULT, 100)
    if SlotType.DURABLE.value in required_slot_types:
        resolved_config = ensure_slot_config(resolved_config, SlotType.DURABLE, 1000)

    # Raise for any required custom slot types that are missing from the config
    configured_keys = {
        key.value if isinstance(key, SlotType) else key for key in resolved_config
    }
    missing = required_slot_types - configured_keys
    if missing:
        formatted = ", ".join(sorted(missing))
        raise ValueError(
            f"Worker is missing slot config for required slot type(s): {formatted}. "
            "Please provide a slot_config entry for each custom slot type used by your workflows."
        )

    if not resolved_config:
        resolved_config = {SlotType.DEFAULT: 100}

    return resolved_config



# coding: utf-8

"""
Hatchet API

The Hatchet API

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import json
from enum import Enum
from typing_extensions import Self


class V1WorkflowType(str, Enum):
    """
    V1WorkflowType
    """

    """
    allowed enum values
    """
    DAG = "DAG"
    TASK = "TASK"

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of V1WorkflowType from a JSON string"""
        return cls(json.loads(json_str))



"""Legacy dual-worker orchestration for pre-slot-config engines.

When connected to an older Hatchet engine that does not support multiple slot types,
this module provides the old worker start flow which:
  - Splits tasks into durable and non-durable registries
  - Spawns separate action listener processes for each
  - Creates separate action runners for each
  - Monitors health of both processes
"""

from __future__ import annotations

import asyncio
import multiprocessing.context
import os
import sys
from multiprocessing import Queue
from typing import TYPE_CHECKING, Any

from hatchet_sdk.deprecated.action_listener_process import (
    legacy_worker_action_listener_process,
)
from hatchet_sdk.logger import logger
from hatchet_sdk.runnables.action import Action
from hatchet_sdk.runnables.contextvars import task_count
from hatchet_sdk.runnables.task import Task
from hatchet_sdk.utils.typing import STOP_LOOP_TYPE
from hatchet_sdk.worker.action_listener_process import ActionEvent
from hatchet_sdk.worker.runner.run_loop_manager import WorkerActionRunLoopManager

if TYPE_CHECKING:
    from hatchet_sdk.worker.worker import Worker


async def legacy_aio_start(worker: Worker) -> None:
    """Start the worker using the legacy dual-worker architecture.

    This is the old _aio_start flow that splits durable and non-durable tasks
    into separate processes, for engines that don't understand slot_config.
    """
    from hatchet_sdk.exceptions import LifespanSetupError
    from hatchet_sdk.worker.worker import WorkerStatus

    main_pid = os.getpid()

    logger.info("------------------------------------------")
    logger.info("STARTING HATCHET (legacy mode)...")
    logger.debug(f"worker runtime starting on PID: {main_pid}")

    worker._status = WorkerStatus.STARTING

    if len(worker.action_registry.keys()) == 0:
        raise ValueError(
            "no actions registered, register workflows or actions before starting worker"
        )

    # Split the unified action registry into durable/non-durable
    durable_action_registry: dict[str, Task[Any, Any]] = {}
    non_durable_action_registry: dict[str, Task[Any, Any]] = {}

    for action_name, task in worker.action_registry.items():
        if task.is_durable:
            durable_action_registry[action_name] = task
        else:
            non_durable_action_registry[action_name] = task

    has_any_non_durable = len(non_durable_action_registry) > 0
    has_any_durable = len(durable_action_registry) > 0

    # Create separate queues for durable workers
    durable_action_queue: Queue[Action | STOP_LOOP_TYPE] = worker.ctx.Queue()
    durable_event_queue: Queue[ActionEvent] = worker.ctx.Queue()

    lifespan_context = None
    if worker.lifespan:
        try:
            lifespan_context = await worker._setup_lifespan()
        except LifespanSetupError as e:
            logger.exception("lifespan setup failed")
            if worker.loop:
                worker.loop.stop()
            raise e

    # Slot conversion: use default and durable from slot_config
    default_slots = worker.slot_config.get("default", 100)
    durable_slots = worker.slot_config.get("durable", 1000)

    durable_action_listener_process = None
    durable_action_runner = None

    if has_any_non_durable:
        worker.action_listener_process = _legacy_start_action_listener(
            worker,
            is_durable=False,
            actions=list(non_durable_action_registry.keys()),
            slots=default_slots,
            action_queue=worker.action_queue,
            event_queue=worker.event_queue,
        )
        worker.action_runner = _legacy_run_action_runner(
            worker,
            name_suffix="",
            action_registry=non_durable_action_registry,
            max_runs=default_slots,
            action_queue=worker.action_queue,
            event_queue=worker.event_queue,
            lifespan_context=lifespan_context,
        )

    if has_any_durable:
        durable_action_listener_process = _legacy_start_action_listener(
            worker,
            is_durable=True,
            actions=list(durable_action_registry.keys()),
            slots=durable_slots,
            action_queue=durable_action_queue,
            event_queue=durable_event_queue,
        )
        durable_action_runner = _legacy_run_action_runner(
            worker,
            name_suffix="_durable",
            action_registry=durable_action_registry,
            max_runs=durable_slots,
            action_queue=durable_action_queue,
            event_queue=durable_event_queue,
            lifespan_context=lifespan_context,
        )

    if worker.loop:
        # Store references for cleanup BEFORE the health check blocks,
        # so they are available when exit_gracefully() runs.
        worker.durable_action_listener_process = durable_action_listener_process
        worker.durable_action_queue = durable_action_queue
        worker.durable_event_queue = durable_event_queue
        worker._legacy_durable_action_runner = durable_action_runner

        worker._lifespan_cleanup_complete = asyncio.Event()
        worker.action_listener_health_check = worker.loop.create_task(
            _legacy_check_listener_health(
                worker,
                durable_action_listener_process,
            )
        )

        await worker.action_listener_health_check

        try:
            await worker._cleanup_lifespan()
        except Exception:
            logger.exception("lifespan cleanup failed")
        finally:
            worker._lifespan_cleanup_complete.set()


def _legacy_start_action_listener(
    worker: Worker,
    is_durable: bool,
    actions: list[str],
    slots: int,
    action_queue: Queue[Any],
    event_queue: Queue[Any],
) -> multiprocessing.context.SpawnProcess:
    try:
        process = worker.ctx.Process(
            target=legacy_worker_action_listener_process,
            args=(
                worker.name + ("_durable" if is_durable else ""),
                actions,
                slots,
                worker.config,
                action_queue,
                event_queue,
                worker.handle_kill,
                worker.client.debug,
                worker.labels,
            ),
        )
        process.start()
        logger.debug(
            f"legacy action listener ({'durable' if is_durable else 'non-durable'}) starting on PID: {process.pid}"
        )
        return process
    except Exception:
        logger.exception("failed to start legacy action listener")
        sys.exit(1)


def _legacy_run_action_runner(
    worker: Worker,
    name_suffix: str,
    action_registry: dict[str, Task[Any, Any]],
    max_runs: int,
    action_queue: Queue[Any],
    event_queue: Queue[Any],
    lifespan_context: Any | None,
) -> WorkerActionRunLoopManager:
    if worker.loop:
        return WorkerActionRunLoopManager(
            worker.name + name_suffix,
            action_registry,
            max_runs,
            worker.config,
            action_queue,
            event_queue,
            worker.loop,
            worker.handle_kill,
            worker.client.debug,
            worker.labels,
            lifespan_context,
        )

    raise RuntimeError("event loop not set, cannot start action runner")


async def _legacy_check_listener_health(
    worker: Worker,
    durable_action_listener_process: multiprocessing.context.SpawnProcess | None,
) -> None:
    from hatchet_sdk.worker.worker import WorkerStatus

    logger.debug("starting legacy action listener health check...")
    try:
        while not worker.killing:
            if (
                not worker.action_listener_process
                and not durable_action_listener_process
            ) or (
                worker.action_listener_process
                and durable_action_listener_process
                and not worker.action_listener_process.is_alive()
                and not durable_action_listener_process.is_alive()
            ):
                logger.debug("child action listener process killed...")
                worker._status = WorkerStatus.UNHEALTHY
                if worker.loop:
                    worker.loop.create_task(worker.exit_gracefully())
                break

            if (
                worker.config.terminate_worker_after_num_tasks
                and task_count.value >= worker.config.terminate_worker_after_num_tasks
            ):
                if worker.loop:
                    worker.loop.create_task(worker.exit_gracefully())
                break

            worker._status = WorkerStatus.HEALTHY
            await asyncio.sleep(1)
    except Exception:
        logger.exception("error checking listener health")



import datetime

from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SDKS(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    UNKNOWN: _ClassVar[SDKS]
    GO: _ClassVar[SDKS]
    PYTHON: _ClassVar[SDKS]
    TYPESCRIPT: _ClassVar[SDKS]
    RUBY: _ClassVar[SDKS]

class ActionType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    START_STEP_RUN: _ClassVar[ActionType]
    CANCEL_STEP_RUN: _ClassVar[ActionType]
    START_GET_GROUP_KEY: _ClassVar[ActionType]

class GroupKeyActionEventType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    GROUP_KEY_EVENT_TYPE_UNKNOWN: _ClassVar[GroupKeyActionEventType]
    GROUP_KEY_EVENT_TYPE_STARTED: _ClassVar[GroupKeyActionEventType]
    GROUP_KEY_EVENT_TYPE_COMPLETED: _ClassVar[GroupKeyActionEventType]
    GROUP_KEY_EVENT_TYPE_FAILED: _ClassVar[GroupKeyActionEventType]

class StepActionEventType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    STEP_EVENT_TYPE_UNKNOWN: _ClassVar[StepActionEventType]
    STEP_EVENT_TYPE_STARTED: _ClassVar[StepActionEventType]
    STEP_EVENT_TYPE_COMPLETED: _ClassVar[StepActionEventType]
    STEP_EVENT_TYPE_FAILED: _ClassVar[StepActionEventType]
    STEP_EVENT_TYPE_ACKNOWLEDGED: _ClassVar[StepActionEventType]

class ResourceType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    RESOURCE_TYPE_UNKNOWN: _ClassVar[ResourceType]
    RESOURCE_TYPE_STEP_RUN: _ClassVar[ResourceType]
    RESOURCE_TYPE_WORKFLOW_RUN: _ClassVar[ResourceType]

class ResourceEventType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    RESOURCE_EVENT_TYPE_UNKNOWN: _ClassVar[ResourceEventType]
    RESOURCE_EVENT_TYPE_STARTED: _ClassVar[ResourceEventType]
    RESOURCE_EVENT_TYPE_COMPLETED: _ClassVar[ResourceEventType]
    RESOURCE_EVENT_TYPE_FAILED: _ClassVar[ResourceEventType]
    RESOURCE_EVENT_TYPE_CANCELLED: _ClassVar[ResourceEventType]
    RESOURCE_EVENT_TYPE_TIMED_OUT: _ClassVar[ResourceEventType]
    RESOURCE_EVENT_TYPE_STREAM: _ClassVar[ResourceEventType]

class WorkflowRunEventType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    WORKFLOW_RUN_EVENT_TYPE_FINISHED: _ClassVar[WorkflowRunEventType]
UNKNOWN: SDKS
GO: SDKS
PYTHON: SDKS
TYPESCRIPT: SDKS
RUBY: SDKS
START_STEP_RUN: ActionType
CANCEL_STEP_RUN: ActionType
START_GET_GROUP_KEY: ActionType
GROUP_KEY_EVENT_TYPE_UNKNOWN: GroupKeyActionEventType
GROUP_KEY_EVENT_TYPE_STARTED: GroupKeyActionEventType
GROUP_KEY_EVENT_TYPE_COMPLETED: GroupKeyActionEventType
GROUP_KEY_EVENT_TYPE_FAILED: GroupKeyActionEventType
STEP_EVENT_TYPE_UNKNOWN: StepActionEventType
STEP_EVENT_TYPE_STARTED: StepActionEventType
STEP_EVENT_TYPE_COMPLETED: StepActionEventType
STEP_EVENT_TYPE_FAILED: StepActionEventType
STEP_EVENT_TYPE_ACKNOWLEDGED: StepActionEventType
RESOURCE_TYPE_UNKNOWN: ResourceType
RESOURCE_TYPE_STEP_RUN: ResourceType
RESOURCE_TYPE_WORKFLOW_RUN: ResourceType
RESOURCE_EVENT_TYPE_UNKNOWN: ResourceEventType
RESOURCE_EVENT_TYPE_STARTED: ResourceEventType
RESOURCE_EVENT_TYPE_COMPLETED: ResourceEventType
RESOURCE_EVENT_TYPE_FAILED: ResourceEventType
RESOURCE_EVENT_TYPE_CANCELLED: ResourceEventType
RESOURCE_EVENT_TYPE_TIMED_OUT: ResourceEventType
RESOURCE_EVENT_TYPE_STREAM: ResourceEventType
WORKFLOW_RUN_EVENT_TYPE_FINISHED: WorkflowRunEventType

class WorkerLabels(_message.Message):
    __slots__ = ("str_value", "int_value")
    STR_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT_VALUE_FIELD_NUMBER: _ClassVar[int]
    str_value: str
    int_value: int
    def __init__(self, str_value: _Optional[str] = ..., int_value: _Optional[int] = ...) -> None: ...

class RuntimeInfo(_message.Message):
    __slots__ = ("sdk_version", "language", "language_version", "os", "extra")
    SDK_VERSION_FIELD_NUMBER: _ClassVar[int]
    LANGUAGE_FIELD_NUMBER: _ClassVar[int]
    LANGUAGE_VERSION_FIELD_NUMBER: _ClassVar[int]
    OS_FIELD_NUMBER: _ClassVar[int]
    EXTRA_FIELD_NUMBER: _ClassVar[int]
    sdk_version: str
    language: SDKS
    language_version: str
    os: str
    extra: str
    def __init__(self, sdk_version: _Optional[str] = ..., language: _Optional[_Union[SDKS, str]] = ..., language_version: _Optional[str] = ..., os: _Optional[str] = ..., extra: _Optional[str] = ...) -> None: ...

class WorkerRegisterRequest(_message.Message):
    __slots__ = ("worker_name", "actions", "services", "slots", "labels", "webhook_id", "runtime_info", "slot_config")
    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: WorkerLabels
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[WorkerLabels, _Mapping]] = ...) -> None: ...
    class SlotConfigEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: int
        def __init__(self, key: _Optional[str] = ..., value: _Optional[int] = ...) -> None: ...
    WORKER_NAME_FIELD_NUMBER: _ClassVar[int]
    ACTIONS_FIELD_NUMBER: _ClassVar[int]
    SERVICES_FIELD_NUMBER: _ClassVar[int]
    SLOTS_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    WEBHOOK_ID_FIELD_NUMBER: _ClassVar[int]
    RUNTIME_INFO_FIELD_NUMBER: _ClassVar[int]
    SLOT_CONFIG_FIELD_NUMBER: _ClassVar[int]
    worker_name: str
    actions: _containers.RepeatedScalarFieldContainer[str]
    services: _containers.RepeatedScalarFieldContainer[str]
    slots: int
    labels: _containers.MessageMap[str, WorkerLabels]
    webhook_id: str
    runtime_info: RuntimeInfo
    slot_config: _containers.ScalarMap[str, int]
    def __init__(self, worker_name: _Optional[str] = ..., actions: _Optional[_Iterable[str]] = ..., services: _Optional[_Iterable[str]] = ..., slots: _Optional[int] = ..., labels: _Optional[_Mapping[str, WorkerLabels]] = ..., webhook_id: _Optional[str] = ..., runtime_info: _Optional[_Union[RuntimeInfo, _Mapping]] = ..., slot_config: _Optional[_Mapping[str, int]] = ...) -> None: ...

class WorkerRegisterResponse(_message.Message):
    __slots__ = ("tenant_id", "worker_id", "worker_name")
    TENANT_ID_FIELD_NUMBER: _ClassVar[int]
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    WORKER_NAME_FIELD_NUMBER: _ClassVar[int]
    tenant_id: str
    worker_id: str
    worker_name: str
    def __init__(self, tenant_id: _Optional[str] = ..., worker_id: _Optional[str] = ..., worker_name: _Optional[str] = ...) -> None: ...

class UpsertWorkerLabelsRequest(_message.Message):
    __slots__ = ("worker_id", "labels")
    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: WorkerLabels
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[WorkerLabels, _Mapping]] = ...) -> None: ...
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    worker_id: str
    labels: _containers.MessageMap[str, WorkerLabels]
    def __init__(self, worker_id: _Optional[str] = ..., labels: _Optional[_Mapping[str, WorkerLabels]] = ...) -> None: ...

class UpsertWorkerLabelsResponse(_message.Message):
    __slots__ = ("tenant_id", "worker_id")
    TENANT_ID_FIELD_NUMBER: _ClassVar[int]
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    tenant_id: str
    worker_id: str
    def __init__(self, tenant_id: _Optional[str] = ..., worker_id: _Optional[str] = ...) -> None: ...

class AssignedAction(_message.Message):
    __slots__ = ("tenant_id", "workflow_run_id", "get_group_key_run_id", "job_id", "job_name", "job_run_id", "task_id", "task_run_external_id", "action_id", "action_type", "action_payload", "task_name", "retry_count", "additional_metadata", "child_workflow_index", "child_workflow_key", "parent_workflow_run_id", "priority", "workflow_id", "workflow_version_id")
    TENANT_ID_FIELD_NUMBER: _ClassVar[int]
    WORKFLOW_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    GET_GROUP_KEY_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    JOB_NAME_FIELD_NUMBER: _ClassVar[int]
    JOB_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    TASK_RUN_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    ACTION_ID_FIELD_NUMBER: _ClassVar[int]
    ACTION_TYPE_FIELD_NUMBER: _ClassVar[int]
    ACTION_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    TASK_NAME_FIELD_NUMBER: _ClassVar[int]
    RETRY_COUNT_FIELD_NUMBER: _ClassVar[int]
    ADDITIONAL_METADATA_FIELD_NUMBER: _ClassVar[int]
    CHILD_WORKFLOW_INDEX_FIELD_NUMBER: _ClassVar[int]
    CHILD_WORKFLOW_KEY_FIELD_NUMBER: _ClassVar[int]
    PARENT_WORKFLOW_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    PRIORITY_FIELD_NUMBER: _ClassVar[int]
    WORKFLOW_ID_FIELD_NUMBER: _ClassVar[int]
    WORKFLOW_VERSION_ID_FIELD_NUMBER: _ClassVar[int]
    tenant_id: str
    workflow_run_id: str
    get_group_key_run_id: str
    job_id: str
    job_name: str
    job_run_id: str
    task_id: str
    task_run_external_id: str
    action_id: str
    action_type: ActionType
    action_payload: str
    task_name: str
    retry_count: int
    additional_metadata: str
    child_workflow_index: int
    child_workflow_key: str
    parent_workflow_run_id: str
    priority: int
    workflow_id: str
    workflow_version_id: str
    def __init__(self, tenant_id: _Optional[str] = ..., workflow_run_id: _Optional[str] = ..., get_group_key_run_id: _Optional[str] = ..., job_id: _Optional[str] = ..., job_name: _Optional[str] = ..., job_run_id: _Optional[str] = ..., task_id: _Optional[str] = ..., task_run_external_id: _Optional[str] = ..., action_id: _Optional[str] = ..., action_type: _Optional[_Union[ActionType, str]] = ..., action_payload: _Optional[str] = ..., task_name: _Optional[str] = ..., retry_count: _Optional[int] = ..., additional_metadata: _Optional[str] = ..., child_workflow_index: _Optional[int] = ..., child_workflow_key: _Optional[str] = ..., parent_workflow_run_id: _Optional[str] = ..., priority: _Optional[int] = ..., workflow_id: _Optional[str] = ..., workflow_version_id: _Optional[str] = ...) -> None: ...

class WorkerListenRequest(_message.Message):
    __slots__ = ("worker_id",)
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    worker_id: str
    def __init__(self, worker_id: _Optional[str] = ...) -> None: ...

class WorkerUnsubscribeRequest(_message.Message):
    __slots__ = ("worker_id",)
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    worker_id: str
    def __init__(self, worker_id: _Optional[str] = ...) -> None: ...

class WorkerUnsubscribeResponse(_message.Message):
    __slots__ = ("tenant_id", "worker_id")
    TENANT_ID_FIELD_NUMBER: _ClassVar[int]
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    tenant_id: str
    worker_id: str
    def __init__(self, tenant_id: _Optional[str] = ..., worker_id: _Optional[str] = ...) -> None: ...

class GroupKeyActionEvent(_message.Message):
    __slots__ = ("worker_id", "workflow_run_id", "get_group_key_run_id", "action_id", "event_timestamp", "event_type", "event_payload")
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    WORKFLOW_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    GET_GROUP_KEY_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    ACTION_ID_FIELD_NUMBER: _ClassVar[int]
    EVENT_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    EVENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    EVENT_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    worker_id: str
    workflow_run_id: str
    get_group_key_run_id: str
    action_id: str
    event_timestamp: _timestamp_pb2.Timestamp
    event_type: GroupKeyActionEventType
    event_payload: str
    def __init__(self, worker_id: _Optional[str] = ..., workflow_run_id: _Optional[str] = ..., get_group_key_run_id: _Optional[str] = ..., action_id: _Optional[str] = ..., event_timestamp: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., event_type: _Optional[_Union[GroupKeyActionEventType, str]] = ..., event_payload: _Optional[str] = ...) -> None: ...

class StepActionEvent(_message.Message):
    __slots__ = ("worker_id", "job_id", "job_run_id", "task_id", "task_run_external_id", "action_id", "event_timestamp", "event_type", "event_payload", "retry_count", "should_not_retry")
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    JOB_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    TASK_RUN_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    ACTION_ID_FIELD_NUMBER: _ClassVar[int]
    EVENT_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    EVENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    EVENT_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    RETRY_COUNT_FIELD_NUMBER: _ClassVar[int]
    SHOULD_NOT_RETRY_FIELD_NUMBER: _ClassVar[int]
    worker_id: str
    job_id: str
    job_run_id: str
    task_id: str
    task_run_external_id: str
    action_id: str
    event_timestamp: _timestamp_pb2.Timestamp
    event_type: StepActionEventType
    event_payload: str
    retry_count: int
    should_not_retry: bool
    def __init__(self, worker_id: _Optional[str] = ..., job_id: _Optional[str] = ..., job_run_id: _Optional[str] = ..., task_id: _Optional[str] = ..., task_run_external_id: _Optional[str] = ..., action_id: _Optional[str] = ..., event_timestamp: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., event_type: _Optional[_Union[StepActionEventType, str]] = ..., event_payload: _Optional[str] = ..., retry_count: _Optional[int] = ..., should_not_retry: bool = ...) -> None: ...

class ActionEventResponse(_message.Message):
    __slots__ = ("tenant_id", "worker_id")
    TENANT_ID_FIELD_NUMBER: _ClassVar[int]
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    tenant_id: str
    worker_id: str
    def __init__(self, tenant_id: _Optional[str] = ..., worker_id: _Optional[str] = ...) -> None: ...

class SubscribeToWorkflowEventsRequest(_message.Message):
    __slots__ = ("workflow_run_id", "additional_meta_key", "additional_meta_value")
    WORKFLOW_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    ADDITIONAL_META_KEY_FIELD_NUMBER: _ClassVar[int]
    ADDITIONAL_META_VALUE_FIELD_NUMBER: _ClassVar[int]
    workflow_run_id: str
    additional_meta_key: str
    additional_meta_value: str
    def __init__(self, workflow_run_id: _Optional[str] = ..., additional_meta_key: _Optional[str] = ..., additional_meta_value: _Optional[str] = ...) -> None: ...

class SubscribeToWorkflowRunsRequest(_message.Message):
    __slots__ = ("workflow_run_id",)
    WORKFLOW_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    workflow_run_id: str
    def __init__(self, workflow_run_id: _Optional[str] = ...) -> None: ...

class WorkflowEvent(_message.Message):
    __slots__ = ("workflow_run_id", "resource_type", "event_type", "resource_id", "event_timestamp", "event_payload", "hangup", "task_retries", "retry_count", "event_index")
    WORKFLOW_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_TYPE_FIELD_NUMBER: _ClassVar[int]
    EVENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_ID_FIELD_NUMBER: _ClassVar[int]
    EVENT_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    EVENT_PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    HANGUP_FIELD_NUMBER: _ClassVar[int]
    TASK_RETRIES_FIELD_NUMBER: _ClassVar[int]
    RETRY_COUNT_FIELD_NUMBER: _ClassVar[int]
    EVENT_INDEX_FIELD_NUMBER: _ClassVar[int]
    workflow_run_id: str
    resource_type: ResourceType
    event_type: ResourceEventType
    resource_id: str
    event_timestamp: _timestamp_pb2.Timestamp
    event_payload: str
    hangup: bool
    task_retries: int
    retry_count: int
    event_index: int
    def __init__(self, workflow_run_id: _Optional[str] = ..., resource_type: _Optional[_Union[ResourceType, str]] = ..., event_type: _Optional[_Union[ResourceEventType, str]] = ..., resource_id: _Optional[str] = ..., event_timestamp: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., event_payload: _Optional[str] = ..., hangup: bool = ..., task_retries: _Optional[int] = ..., retry_count: _Optional[int] = ..., event_index: _Optional[int] = ...) -> None: ...

class WorkflowRunEvent(_message.Message):
    __slots__ = ("workflow_run_id", "event_type", "event_timestamp", "results")
    WORKFLOW_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    EVENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    EVENT_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    RESULTS_FIELD_NUMBER: _ClassVar[int]
    workflow_run_id: str
    event_type: WorkflowRunEventType
    event_timestamp: _timestamp_pb2.Timestamp
    results: _containers.RepeatedCompositeFieldContainer[StepRunResult]
    def __init__(self, workflow_run_id: _Optional[str] = ..., event_type: _Optional[_Union[WorkflowRunEventType, str]] = ..., event_timestamp: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., results: _Optional[_Iterable[_Union[StepRunResult, _Mapping]]] = ...) -> None: ...

class StepRunResult(_message.Message):
    __slots__ = ("task_run_external_id", "task_name", "job_run_id", "error", "output")
    TASK_RUN_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    TASK_NAME_FIELD_NUMBER: _ClassVar[int]
    JOB_RUN_ID_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_FIELD_NUMBER: _ClassVar[int]
    task_run_external_id: str
    task_name: str
    job_run_id: str
    error: str
    output: str
    def __init__(self, task_run_external_id: _Optional[str] = ..., task_name: _Optional[str] = ..., job_run_id: _Optional[str] = ..., error: _Optional[str] = ..., output: _Optional[str] = ...) -> None: ...

class OverridesData(_message.Message):
    __slots__ = ("task_run_external_id", "path", "value", "caller_filename")
    TASK_RUN_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    CALLER_FILENAME_FIELD_NUMBER: _ClassVar[int]
    task_run_external_id: str
    path: str
    value: str
    caller_filename: str
    def __init__(self, task_run_external_id: _Optional[str] = ..., path: _Optional[str] = ..., value: _Optional[str] = ..., caller_filename: _Optional[str] = ...) -> None: ...

class OverridesDataResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class HeartbeatRequest(_message.Message):
    __slots__ = ("worker_id", "heartbeat_at")
    WORKER_ID_FIELD_NUMBER: _ClassVar[int]
    HEARTBEAT_AT_FIELD_NUMBER: _ClassVar[int]
    worker_id: str
    heartbeat_at: _timestamp_pb2.Timestamp
    def __init__(self, worker_id: _Optional[str] = ..., heartbeat_at: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class HeartbeatResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class RefreshTimeoutRequest(_message.Message):
    __slots__ = ("task_run_external_id", "increment_timeout_by")
    TASK_RUN_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    INCREMENT_TIMEOUT_BY_FIELD_NUMBER: _ClassVar[int]
    task_run_external_id: str
    increment_timeout_by: str
    def __init__(self, task_run_external_id: _Optional[str] = ..., increment_timeout_by: _Optional[str] = ...) -> None: ...

class RefreshTimeoutResponse(_message.Message):
    __slots__ = ("timeout_at",)
    TIMEOUT_AT_FIELD_NUMBER: _ClassVar[int]
    timeout_at: _timestamp_pb2.Timestamp
    def __init__(self, timeout_at: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class ReleaseSlotRequest(_message.Message):
    __slots__ = ("task_run_external_id",)
    TASK_RUN_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    task_run_external_id: str
    def __init__(self, task_run_external_id: _Optional[str] = ...) -> None: ...

class ReleaseSlotResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetVersionRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetVersionResponse(_message.Message):
    __slots__ = ("version",)
    VERSION_FIELD_NUMBER: _ClassVar[int]
    version: str
    def __init__(self, version: _Optional[str] = ...) -> None: ...



# coding: utf-8

"""
Hatchet API

The Hatchet API

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from typing_extensions import Annotated
from hatchet_sdk.clients.rest.models.api_resource_meta import APIResourceMeta
from hatchet_sdk.clients.rest.models.v1_task_status import V1TaskStatus
from typing import Optional, Set
from typing_extensions import Self


class V1WorkflowRun(BaseModel):
    """
    V1WorkflowRun
    """  # noqa: E501

    metadata: APIResourceMeta
    status: V1TaskStatus
    started_at: Optional[datetime] = Field(
        default=None,
        description="The timestamp the task run started.",
        alias="startedAt",
    )
    finished_at: Optional[datetime] = Field(
        default=None,
        description="The timestamp the task run finished.",
        alias="finishedAt",
    )
    duration: Optional[StrictInt] = Field(
        default=None, description="The duration of the task run, in milliseconds."
    )
    tenant_id: Annotated[str, Field(min_length=36, strict=True, max_length=36)] = Field(
        description="The ID of the tenant.", alias="tenantId"
    )
    additional_metadata: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Additional metadata for the task run.",
        alias="additionalMetadata",
    )
    display_name: StrictStr = Field(
        description="The display name of the task run.", alias="displayName"
    )
    workflow_id: StrictStr = Field(alias="workflowId")
    output: Dict[str, Any] = Field(
        description="The output of the task run (for the latest run)"
    )
    error_message: Optional[StrictStr] = Field(
        default=None,
        description="The error message of the task run (for the latest run)",
        alias="errorMessage",
    )
    workflow_version_id: Optional[StrictStr] = Field(
        default=None,
        description="The ID of the workflow version.",
        alias="workflowVersionId",
    )
    input: Dict[str, Any] = Field(description="The input of the task run.")
    created_at: Optional[datetime] = Field(
        default=None,
        description="The timestamp the task run was created.",
        alias="createdAt",
    )
    parent_task_external_id: Optional[
        Annotated[str, Field(min_length=36, strict=True, max_length=36)]
    ] = Field(default=None, alias="parentTaskExternalId")
    __properties: ClassVar[List[str]] = [
        "metadata",
        "status",
        "startedAt",
        "finishedAt",
        "duration",
        "tenantId",
        "additionalMetadata",
        "displayName",
        "workflowId",
        "output",
        "errorMessage",
        "workflowVersionId",
        "input",
        "createdAt",
        "parentTaskExternalId",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of V1WorkflowRun from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of metadata
        if self.metadata:
            _dict["metadata"] = self.metadata.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of V1WorkflowRun from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "metadata": (
                    APIResourceMeta.from_dict(obj["metadata"])
                    if obj.get("metadata") is not None
                    else None
                ),
                "status": obj.get("status"),
                "startedAt": obj.get("startedAt"),
                "finishedAt": obj.get("finishedAt"),
                "duration": obj.get("duration"),
                "tenantId": obj.get("tenantId"),
                "additionalMetadata": obj.get("additionalMetadata"),
                "displayName": obj.get("displayName"),
                "workflowId": obj.get("workflowId"),
                "output": obj.get("output"),
                "errorMessage": obj.get("errorMessage"),
                "workflowVersionId": obj.get("workflowVersionId"),
                "input": obj.get("input"),
                "createdAt": obj.get("createdAt"),
                "parentTaskExternalId": obj.get("parentTaskExternalId"),
            }
        )
        return _obj



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, ParamSpec, TypeVar

import grpc
import tenacity

from hatchet_sdk.clients.rest.exceptions import (
    NotFoundException,
    RestTransportError,
    ServiceException,
    TooManyRequestsException,
)
from hatchet_sdk.logger import logger

if TYPE_CHECKING:
    from hatchet_sdk.config import TenacityConfig

P = ParamSpec("P")
R = TypeVar("R")


def tenacity_retry(func: Callable[P, R], config: TenacityConfig) -> Callable[P, R]:
    if config.max_attempts <= 0:
        return func

    def should_retry(ex: BaseException) -> bool:
        return tenacity_should_retry(ex, config)

    return tenacity.retry(
        reraise=True,
        wait=tenacity.wait_exponential_jitter(),
        stop=tenacity.stop_after_attempt(config.max_attempts),
        before_sleep=tenacity_alert_retry,
        retry=tenacity.retry_if_exception(should_retry),
    )(func)


def tenacity_alert_retry(retry_state: tenacity.RetryCallState) -> None:
    """Called between tenacity retries."""
    logger.debug(
        f"retrying {retry_state.fn}: attempt "
        f"{retry_state.attempt_number} ended with: {retry_state.outcome}",
    )


def tenacity_should_retry(
    ex: BaseException, config: TenacityConfig | None = None
) -> bool:
    """Return True when the exception should be retried."""
    if isinstance(ex, ServiceException | NotFoundException):
        return True

    if isinstance(ex, TooManyRequestsException):
        return bool(config and config.retry_429)

    # gRPC errors: retry most, except specific permanent failure codes
    if isinstance(ex, grpc.aio.AioRpcError | grpc.RpcError):
        return ex.code() not in [
            grpc.StatusCode.UNIMPLEMENTED,
            grpc.StatusCode.NOT_FOUND,
            grpc.StatusCode.INVALID_ARGUMENT,
            grpc.StatusCode.ALREADY_EXISTS,
            grpc.StatusCode.UNAUTHENTICATED,
            grpc.StatusCode.PERMISSION_DENIED,
        ]

    # REST transport errors: opt-in retry for configured HTTP methods
    if isinstance(ex, RestTransportError):
        if config is not None and config.retry_transport_errors:
            method = ex.http_method
            if method is not None:
                return method in config.retry_transport_methods
        return False

    return False



from pydantic import BaseModel

from hatchet_sdk.contracts.v1.workflows_pb2 import DesiredWorkerLabels


class DesiredWorkerLabel(BaseModel):
    value: str | int
    required: bool = False
    weight: int | None = None
    comparator: int | None = None


def transform_desired_worker_label(d: DesiredWorkerLabel) -> DesiredWorkerLabels:
    value = d.value
    return DesiredWorkerLabels(
        str_value=value if not isinstance(value, int) else None,
        int_value=value if isinstance(value, int) else None,
        required=d.required,
        weight=d.weight,
        comparator=d.comparator,  # type: ignore[arg-type]
    )



"""Generic time-aware deprecation helper.

Timeline (from a given start date, with configurable windows):
  0 to warn_days:              WARNING logged once per feature
  warn_days to error_days:     ERROR logged once per feature
  after error_days:            raises DeprecationError 1-in-5 calls (20% chance)

Defaults: warn_days=90, error_days=None (error phase disabled unless explicitly set).
"""

import random
from datetime import datetime, timezone

from hatchet_sdk.logger import logger

_DEFAULT_WARN_DAYS = 90

# Tracks which features have already been logged (keyed by feature name).
_already_logged: set[str] = set()


class DeprecationError(Exception):
    """Raised when a deprecation grace period has expired."""


def parse_semver(v: str) -> tuple[int, int, int]:
    """Parse a semver string like ``"v0.78.23"`` into ``(major, minor, patch)``.

    Returns ``(0, 0, 0)`` if parsing fails.
    """
    v = v.lstrip("v").split("-", 1)[0]
    parts = v.split(".")
    if len(parts) != 3:
        return (0, 0, 0)
    try:
        return (int(parts[0]), int(parts[1]), int(parts[2]))
    except ValueError:
        return (0, 0, 0)


def semver_less_than(a: str, b: str) -> bool:
    """Return ``True`` if semver string *a* is strictly less than *b*."""
    return parse_semver(a) < parse_semver(b)


def emit_deprecation_notice(
    feature: str,
    message: str,
    start: datetime,
    *,
    warn_days: int = _DEFAULT_WARN_DAYS,
    error_days: int | None = None,
) -> None:
    """Emit a time-aware deprecation notice.

    Args:
        feature:    A short identifier for the deprecated feature (used for
                    deduplication so each feature only logs once per process).
        message:    The human-readable deprecation message.
        start:      The UTC datetime when the deprecation window began.
        warn_days:  Days after *start* during which a warning is logged (default 90).
        error_days: Days after *start* during which an error is logged.
                    After this window, calls have a 20% chance of raising.
                    If None (default), the error/raise phase is never reached —
                    the notice stays at error-level logging indefinitely.

    Raises:
        DeprecationError: After the error_days window, raised ~20% of the time.
    """
    now = datetime.now(tz=timezone.utc)
    days_since = (now - start).days

    if days_since < warn_days:
        # Phase 1: warning
        if feature not in _already_logged:
            logger.warning(message)
            _already_logged.add(feature)

    elif error_days is None or days_since < error_days:
        # Phase 2: error-level log (indefinite when error_days is None)
        if feature not in _already_logged:
            logger.error(
                f"{message} " "This fallback will be removed soon. Upgrade immediately."
            )
            _already_logged.add(feature)

    else:
        # Phase 3: raise 1-in-5 times
        if feature not in _already_logged:
            logger.error(
                f"{message} "
                "This fallback is no longer supported and will fail intermittently."
            )
            _already_logged.add(feature)

        if random.random() < 0.2:
            raise DeprecationError(f"{feature}: {message}")



# coding: utf-8

"""
Hatchet API

The Hatchet API

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import json
from enum import Enum
from typing_extensions import Self


class V1TaskRunStatus(str, Enum):
    """
    V1TaskRunStatus
    """

    """
    allowed enum values
    """
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of V1TaskRunStatus from a JSON string"""
        return cls(json.loads(json_str))



from __future__ import annotations

import asyncio
from typing import Any

from hatchet_sdk.clients.rest.api.webhook_api import WebhookApi
from hatchet_sdk.clients.rest.api_client import ApiClient
from hatchet_sdk.clients.rest.models.v1_create_webhook_request import (
    V1CreateWebhookRequest,
)
from hatchet_sdk.clients.rest.models.v1_create_webhook_request_api_key import (
    V1CreateWebhookRequestAPIKey,
)
from hatchet_sdk.clients.rest.models.v1_create_webhook_request_base import (
    V1CreateWebhookRequestBase,
)
from hatchet_sdk.clients.rest.models.v1_create_webhook_request_basic_auth import (
    V1CreateWebhookRequestBasicAuth,
)
from hatchet_sdk.clients.rest.models.v1_create_webhook_request_hmac import (
    V1CreateWebhookRequestHMAC,
)
from hatchet_sdk.clients.rest.models.v1_update_webhook_request import (
    V1UpdateWebhookRequest,
)
from hatchet_sdk.clients.rest.models.v1_webhook import V1Webhook
from hatchet_sdk.clients.rest.models.v1_webhook_api_key_auth import V1WebhookAPIKeyAuth
from hatchet_sdk.clients.rest.models.v1_webhook_auth_type import V1WebhookAuthType
from hatchet_sdk.clients.rest.models.v1_webhook_basic_auth import V1WebhookBasicAuth
from hatchet_sdk.clients.rest.models.v1_webhook_hmac_auth import V1WebhookHMACAuth
from hatchet_sdk.clients.rest.models.v1_webhook_list import V1WebhookList
from hatchet_sdk.clients.rest.models.v1_webhook_source_name import V1WebhookSourceName
from hatchet_sdk.clients.rest.tenacity_utils import tenacity_retry
from hatchet_sdk.clients.v1.api_client import BaseRestClient


class CreateWebhookRequest(V1CreateWebhookRequestBase):
    auth: V1WebhookBasicAuth | V1WebhookAPIKeyAuth | V1WebhookHMACAuth

    def _to_api_payload(self) -> V1CreateWebhookRequest:
        payload = self.model_dump(by_alias=True, exclude_none=True)
        payload["auth"] = self.auth.model_dump(by_alias=True)
        request_payload: (
            V1CreateWebhookRequestBasicAuth
            | V1CreateWebhookRequestAPIKey
            | V1CreateWebhookRequestHMAC
            | None
        ) = None
        if isinstance(self.auth, V1WebhookBasicAuth):
            payload["authType"] = V1WebhookAuthType.BASIC
            request_payload = V1CreateWebhookRequestBasicAuth.from_dict(payload)
        elif isinstance(self.auth, V1WebhookAPIKeyAuth):
            payload["authType"] = V1WebhookAuthType.API_KEY
            request_payload = V1CreateWebhookRequestAPIKey.from_dict(payload)
        else:
            payload["authType"] = V1WebhookAuthType.HMAC
            request_payload = V1CreateWebhookRequestHMAC.from_dict(payload)
        if request_payload is None:
            raise ValueError("failed to build create webhook request from payload")
        return V1CreateWebhookRequest(request_payload)


class WebhooksClient(BaseRestClient):
    """
    The webhooks client provides methods for managing incoming webhooks in Hatchet.

    Webhooks allow external systems to trigger Hatchet workflows by sending HTTP
    requests to dedicated endpoints. This enables real-time integration with
    third-party services like GitHub, Stripe, Slack, or any system that can send
    webhook events.
    """

    def _wa(self, client: ApiClient) -> WebhookApi:
        return WebhookApi(client)

    async def aio_list(
        self,
        limit: int | None = None,
        offset: int | None = None,
        webhook_names: list[str] | None = None,
        source_names: list[V1WebhookSourceName] | None = None,
    ) -> V1WebhookList:
        return await asyncio.to_thread(
            self.list, limit, offset, webhook_names, source_names
        )

    def list(
        self,
        limit: int | None = None,
        offset: int | None = None,
        webhook_names: list[str] | None = None,
        source_names: list[V1WebhookSourceName] | None = None,
    ) -> V1WebhookList:
        with self.client() as client:
            v1_webhook_list = tenacity_retry(
                self._wa(client).v1_webhook_list, self.client_config.tenacity
            )
            return v1_webhook_list(
                tenant=self.tenant_id,
                limit=limit,
                offset=offset,
                webhook_names=webhook_names,
                source_names=source_names,
            )

    def get(self, webhook_name: str) -> V1Webhook:
        with self.client() as client:
            v1_webhook_get = tenacity_retry(
                self._wa(client).v1_webhook_get, self.client_config.tenacity
            )
            return v1_webhook_get(
                tenant=self.tenant_id,
                v1_webhook=webhook_name,
            )

    async def aio_get(self, webhook_name: str) -> V1Webhook:
        return await asyncio.to_thread(self.get, webhook_name)

    def create(
        self,
        source_name: V1WebhookSourceName,
        name: str,
        event_key_expression: str,
        auth: V1WebhookBasicAuth | V1WebhookAPIKeyAuth | V1WebhookHMACAuth,
        scope_expression: str | None = None,
        static_payload: dict[str, Any] | None = None,
    ) -> V1Webhook:
        validated_payload = CreateWebhookRequest(
            sourceName=source_name,
            name=name,
            eventKeyExpression=event_key_expression,
            scopeExpression=scope_expression,
            staticPayload=static_payload,
            auth=auth,
        )
        with self.client() as client:
            return self._wa(client).v1_webhook_create(
                tenant=self.tenant_id,
                v1_create_webhook_request=validated_payload._to_api_payload(),
            )

    async def aio_create(
        self,
        source_name: V1WebhookSourceName,
        name: str,
        event_key_expression: str,
        auth: V1WebhookBasicAuth | V1WebhookAPIKeyAuth | V1WebhookHMACAuth,
        scope_expression: str | None = None,
        static_payload: dict[str, Any] | None = None,
    ) -> V1Webhook:
        return await asyncio.to_thread(
            self.create,
            source_name,
            name,
            event_key_expression,
            auth,
            scope_expression,
            static_payload,
        )

    def update(
        self,
        webhook_name: str,
        event_key_expression: str | None = None,
        scope_expression: str | None = None,
        static_payload: dict[str, Any] | None = None,
    ) -> V1Webhook:
        with self.client() as client:
            return self._wa(client).v1_webhook_update(
                tenant=self.tenant_id,
                v1_webhook=webhook_name,
                v1_update_webhook_request=V1UpdateWebhookRequest(
                    eventKeyExpression=event_key_expression,
                    scopeExpression=scope_expression,
                    staticPayload=static_payload,
                ),
            )

    async def aio_update(
        self,
        webhook_name: str,
        event_key_expression: str | None = None,
        scope_expression: str | None = None,
        static_payload: dict[str, Any] | None = None,
    ) -> V1Webhook:
        return await asyncio.to_thread(
            self.update,
            webhook_name,
            event_key_expression,
            scope_expression,
            static_payload,
        )

    def delete(self, webhook_name: str) -> V1Webhook:
        with self.client() as client:
            return self._wa(client).v1_webhook_delete(
                tenant=self.tenant_id,
                v1_webhook=webhook_name,
            )

    async def aio_delete(self, webhook_name: str) -> V1Webhook:
        return await asyncio.to_thread(self.delete, webhook_name)



import asyncio
import multiprocessing
import multiprocessing.context
import os
import signal
import sys
from collections.abc import AsyncGenerator, Callable
from contextlib import AsyncExitStack, asynccontextmanager, suppress
from dataclasses import dataclass, field
from enum import Enum
from multiprocessing import Queue
from multiprocessing.process import BaseProcess
from types import FrameType
from typing import Any, TypeVar
from warnings import warn

import grpc

from hatchet_sdk.client import Client
from hatchet_sdk.config import ClientConfig
from hatchet_sdk.contracts.v1.workflows_pb2 import CreateWorkflowVersionRequest
from hatchet_sdk.deprecated.deprecation import semver_less_than
from hatchet_sdk.deprecated.worker import legacy_aio_start
from hatchet_sdk.exceptions import LifespanSetupError, LoopAlreadyRunningError
from hatchet_sdk.logger import logger
from hatchet_sdk.runnables.action import Action
from hatchet_sdk.runnables.contextvars import task_count
from hatchet_sdk.runnables.task import Task
from hatchet_sdk.runnables.workflow import BaseWorkflow
from hatchet_sdk.utils.typing import STOP_LOOP_TYPE
from hatchet_sdk.worker.action_listener_process import (
    ActionEvent,
    worker_action_listener_process,
)
from hatchet_sdk.worker.runner.run_loop_manager import WorkerActionRunLoopManager

T = TypeVar("T")


class WorkerStatus(Enum):
    INITIALIZED = 1
    STARTING = 2
    HEALTHY = 3
    UNHEALTHY = 4


@dataclass
class WorkerStartOptions:
    loop: asyncio.AbstractEventLoop | None = field(default=None)


LifespanGenerator = AsyncGenerator[Any, Any]
LifespanFn = Callable[[], LifespanGenerator]


@asynccontextmanager
async def _create_async_context_manager(
    gen: LifespanGenerator,
) -> AsyncGenerator[None, None]:
    try:
        yield
    finally:
        with suppress(StopAsyncIteration):
            await anext(gen)


class Worker:
    def __init__(
        self,
        name: str,
        config: ClientConfig,
        slot_config: dict[str, int],
        labels: dict[str, str | int] | None = None,
        debug: bool = False,
        owned_loop: bool = True,
        handle_kill: bool = True,
        workflows: list[BaseWorkflow[Any]] | None = None,
        lifespan: LifespanFn | None = None,
    ) -> None:
        self.config = config
        self.name = self.config.apply_namespace(name)
        self.slot_config = slot_config
        self._slots = slot_config.get("default", 0)
        self._durable_slots = slot_config.get("durable", 0)
        self.debug = debug
        self.labels = labels or {}
        self.handle_kill = handle_kill
        self.owned_loop = owned_loop

        self.action_registry: dict[str, Task[Any, Any]] = {}

        self.killing: bool = False
        self._status: WorkerStatus = WorkerStatus.INITIALIZED

        self.action_listener_process: BaseProcess | None = None
        self.durable_action_listener_process: BaseProcess | None = None

        self.action_listener_health_check: asyncio.Task[None]

        self.action_runner: WorkerActionRunLoopManager | None = None
        self._legacy_durable_action_runner: WorkerActionRunLoopManager | None = None

        self.ctx = multiprocessing.get_context("spawn")

        self.action_queue: Queue[Action | STOP_LOOP_TYPE] = self.ctx.Queue()
        self.event_queue: Queue[ActionEvent] = self.ctx.Queue()
        self.durable_action_queue: Queue[Action | STOP_LOOP_TYPE] | None = None
        self.durable_event_queue: Queue[ActionEvent] | None = None

        self.loop: asyncio.AbstractEventLoop | None = None

        self.client = Client(config=self.config, debug=self.debug)

        self._setup_signal_handlers()

        self.lifespan = lifespan
        self.lifespan_stack: AsyncExitStack | None = None
        self._lifespan_cleanup_complete: asyncio.Event | None = None

        self.register_workflows(workflows or [])

    def register_workflow_from_opts(self, opts: CreateWorkflowVersionRequest) -> None:
        try:
            self.client.admin.put_workflow(opts)
        except Exception:
            logger.exception(f"failed to register workflow: {opts.name}")
            sys.exit(1)

    def register_workflow(self, workflow: BaseWorkflow[Any]) -> None:
        if not workflow.tasks:
            msg = f"failed to register workflow: {workflow.name}. Workflows must have at least one task registered before registering"
            raise ValueError(msg)

        try:
            self.client.admin.put_workflow(workflow.to_proto())
        except Exception:
            logger.exception(f"failed to register workflow: {workflow.name}")
            sys.exit(1)

        for step in workflow.tasks:
            action_name = workflow._create_action_name(step)

            self.action_registry[action_name] = step

    def register_workflows(self, workflows: list[BaseWorkflow[Any]]) -> None:
        for workflow in workflows:
            self.register_workflow(workflow)

    @property
    def status(self) -> WorkerStatus:
        return self._status

    @property
    def slots(self) -> int:
        warn(
            "Worker.slots is deprecated; use slot_config['default'] instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._slots

    @property
    def durable_slots(self) -> int:
        warn(
            "Worker.durable_slots is deprecated; use slot_config['durable'] instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._durable_slots

    def _setup_loop(self) -> None:
        try:
            asyncio.get_running_loop()
            raise LoopAlreadyRunningError(
                "An event loop is already running. This worker requires its own dedicated event loop. "
                "Make sure you're not using asyncio.run() or other loop-creating functions in the main thread."
            )
        except RuntimeError:
            pass

        logger.debug("creating new event loop")
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

    def start(self, options: WorkerStartOptions = WorkerStartOptions()) -> None:
        if not self.action_registry:
            raise ValueError(
                "no actions registered, register workflows before starting worker"
            )

        if options.loop is not None:
            warn(
                "Passing a custom event loop is deprecated and will be removed in the future. This option no longer has any effect",
                DeprecationWarning,
                stacklevel=1,
            )

        self._setup_loop()

        if not self.loop:
            raise RuntimeError("event loop not set, cannot start worker")

        asyncio.run_coroutine_threadsafe(self._aio_start(), self.loop)

        # start the loop and wait until its closed
        self.loop.run_forever()

        if self.handle_kill:
            sys.exit(0)

    # Minimum engine version that supports multiple slot types.
    _MIN_SLOT_CONFIG_VERSION = "v0.78.23"

    def _emit_legacy_deprecation(self) -> None:
        from datetime import datetime, timezone

        from hatchet_sdk.deprecated.deprecation import emit_deprecation_notice

        emit_deprecation_notice(
            feature="legacy-engine",
            message=(
                "Connected to an older Hatchet engine that does not support "
                "multiple slot types. Falling back to legacy worker registration. "
                "Please upgrade your Hatchet engine to the latest version."
            ),
            start=datetime(2026, 2, 12, tzinfo=timezone.utc),
            error_days=180,
        )

    async def _check_engine_version(self) -> str | None:
        """Returns the engine version string, or None if engine is legacy (pre-slot-config).

        Compares the engine's semantic version against the minimum required
        version for slot_config support. Returns the version string for modern
        engines so callers can branch on specific versions.
        """

        try:
            version = await self.client.dispatcher.get_version()

            # Empty version or older than minimum → legacy
            if not version or semver_less_than(version, self._MIN_SLOT_CONFIG_VERSION):
                self._emit_legacy_deprecation()
                return None

            return version  # new engine
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.UNIMPLEMENTED:
                self._emit_legacy_deprecation()
                return None  # old engine
            raise

    async def _aio_start(self) -> None:
        main_pid = os.getpid()

        logger.info("------------------------------------------")
        logger.info("STARTING HATCHET...")
        logger.debug(f"worker runtime starting on PID: {main_pid}")

        self._status = WorkerStatus.STARTING

        if len(self.action_registry.keys()) == 0:
            raise ValueError(
                "no actions registered, register workflows or actions before starting worker"
            )

        # Check engine version and fall back to legacy dual-worker mode if needed
        engine_version = await self._check_engine_version()
        if engine_version is None:
            await legacy_aio_start(self)
            return

        lifespan_context = None
        if self.lifespan:
            try:
                lifespan_context = await self._setup_lifespan()
            except LifespanSetupError as e:
                logger.exception("lifespan setup failed")
                if self.loop:
                    self.loop.stop()
                raise e

        # Healthcheck server is started inside the spawned action-listener process
        # (non-durable preferred) to avoid being affected by the main worker loop.
        healthcheck_port = self.config.healthcheck.port
        enable_health_server = self.config.healthcheck.enabled

        self.action_listener_process = self._start_action_listener(
            enable_health_server=enable_health_server,
            healthcheck_port=healthcheck_port,
        )
        self.action_runner = self._run_action_runner(lifespan_context=lifespan_context)

        if self.loop:
            self._lifespan_cleanup_complete = asyncio.Event()
            self.action_listener_health_check = self.loop.create_task(
                self._check_listener_health()
            )

            await self.action_listener_health_check

            try:
                await self._cleanup_lifespan()
            except LifespanSetupError:
                logger.exception("lifespan cleanup failed")
            finally:
                self._lifespan_cleanup_complete.set()

    def _run_action_runner(
        self, lifespan_context: Any | None
    ) -> WorkerActionRunLoopManager:
        # Retrieve the shared queue
        if self.loop:
            return WorkerActionRunLoopManager(
                self.name,
                self.action_registry,
                sum(self.slot_config.values()),
                self.config,
                self.action_queue,
                self.event_queue,
                self.loop,
                self.handle_kill,
                self.client.debug,
                self.labels,
                lifespan_context,
            )

        raise RuntimeError("event loop not set, cannot start action runner")

    async def _setup_lifespan(self) -> Any:
        if self.lifespan is None:
            return None

        self.lifespan_stack = AsyncExitStack()

        try:
            lifespan_gen = self.lifespan()
            context = await anext(lifespan_gen)
            await self.lifespan_stack.enter_async_context(
                _create_async_context_manager(lifespan_gen)
            )
            return context
        except StopAsyncIteration:
            return None
        except Exception as e:
            raise LifespanSetupError("An error occurred during lifespan setup") from e

    async def _cleanup_lifespan(self) -> None:
        try:
            if self.lifespan_stack is not None:
                await self.lifespan_stack.aclose()
        except Exception as e:
            logger.exception("error during lifespan cleanup")
            raise LifespanSetupError("An error occurred during lifespan cleanup") from e

    def _start_action_listener(
        self,
        *,
        enable_health_server: bool = False,
        healthcheck_port: int = 8001,
    ) -> multiprocessing.context.SpawnProcess:
        try:
            process = self.ctx.Process(
                target=worker_action_listener_process,
                args=(
                    self.name,
                    list(self.action_registry.keys()),
                    self.slot_config,
                    self.config,
                    self.action_queue,
                    self.event_queue,
                    self.handle_kill,
                    self.client.debug,
                    self.labels,
                ),
            )
            process.start()
            logger.debug(f"action listener starting on PID: {process.pid}")

            return process
        except Exception:
            logger.exception("failed to start action listener")
            sys.exit(1)

    async def _check_listener_health(self) -> None:
        logger.debug("starting action listener health check...")
        try:
            while not self.killing:
                if (
                    not self.action_listener_process
                    or not self.action_listener_process.is_alive()
                ):
                    logger.debug("child action listener process killed...")
                    self._status = WorkerStatus.UNHEALTHY
                    if self.loop:
                        self.loop.create_task(self.exit_gracefully())
                    break

                if (
                    self.config.terminate_worker_after_num_tasks
                    and task_count.value >= self.config.terminate_worker_after_num_tasks
                ):
                    if self.loop:
                        self.loop.create_task(self.exit_gracefully())
                    break

                self._status = WorkerStatus.HEALTHY
                await asyncio.sleep(1)
        except Exception:
            logger.exception("error checking listener health")

    def _setup_signal_handlers(self) -> None:
        signal.signal(
            signal.SIGTERM,
            (
                self._handle_force_quit_signal
                if self.config.force_shutdown_on_shutdown_signal
                else self._handle_exit_signal
            ),
        )
        signal.signal(
            signal.SIGINT,
            (
                self._handle_force_quit_signal
                if self.config.force_shutdown_on_shutdown_signal
                else self._handle_exit_signal
            ),
        )
        signal.signal(signal.SIGQUIT, self._handle_force_quit_signal)

    def _handle_exit_signal(self, signum: int, frame: FrameType | None) -> None:
        sig_name = "SIGTERM" if signum == signal.SIGTERM else "SIGINT"
        logger.info(f"received signal {sig_name}...")
        if self.loop:
            self.loop.create_task(self.exit_gracefully())

    def _handle_force_quit_signal(self, signum: int, frame: FrameType | None) -> None:
        signal_received = signal.Signals(signum).name
        logger.info(f"received {signal_received}...")
        if self.loop:
            self.loop.create_task(self._exit_forcefully())

    def _close_queues(self) -> None:
        queues: list[Queue[Any] | None] = [
            self.action_queue,
            self.event_queue,
            self.durable_action_queue,
            self.durable_event_queue,
        ]

        for queue in queues:
            if queue is None:
                continue
            try:
                queue.cancel_join_thread()
                queue.close()
            except Exception:
                continue

    def _terminate_processes(self) -> None:
        for process in [
            self.action_listener_process,
            self.durable_action_listener_process,
        ]:
            if process is not None and process.pid is not None:
                try:
                    if process.is_alive():
                        os.kill(process.pid, signal.SIGQUIT)

                    process.join(timeout=5)

                    if process.is_alive():
                        process.kill()
                        process.join(timeout=1)
                except Exception:
                    pass

    async def _close(self) -> None:
        logger.info(f"closing worker '{self.name}'...")
        self.killing = True

        if self.action_runner is not None:
            self.action_runner.cleanup()

        # Also clean up the durable action runner (legacy mode)
        if self._legacy_durable_action_runner is not None:
            self._legacy_durable_action_runner.cleanup()

        await self.action_listener_health_check

        self._close_queues()

    async def exit_gracefully(self) -> None:
        logger.debug(f"gracefully stopping worker: {self.name}")

        if self.killing:
            return await self._exit_forcefully()

        self.killing = True

        if self.action_runner:
            await self.action_runner.wait_for_tasks()
            await self.action_runner.exit_gracefully()

        # Also clean up the durable action runner (legacy mode)
        if self._legacy_durable_action_runner:
            await self._legacy_durable_action_runner.wait_for_tasks()
            await self._legacy_durable_action_runner.exit_gracefully()

        if self.action_listener_process and self.action_listener_process.is_alive():
            self.action_listener_process.kill()

        if (
            self.durable_action_listener_process
            and self.durable_action_listener_process.is_alive()
        ):
            self.durable_action_listener_process.kill()

        try:
            await self._cleanup_lifespan()
        except LifespanSetupError:
            logger.exception("lifespan cleanup failed")

        await self._close()

        if self._lifespan_cleanup_complete is not None:
            await self._lifespan_cleanup_complete.wait()
        if self.loop and self.owned_loop:
            self.loop.stop()

        logger.info("👋")

    async def _exit_forcefully(self) -> None:
        self.killing = True

        logger.debug(f"forcefully stopping worker: {self.name}")

        self._terminate_processes()

        await self._close()

        if self._lifespan_cleanup_complete is not None:
            try:
                await asyncio.wait_for(
                    self._lifespan_cleanup_complete.wait(), timeout=5.0
                )
            except asyncio.TimeoutError:
                logger.warning("lifespan cleanup timed out during forceful shutdown")

        logger.info("👋")
        sys.exit(1)



"""Legacy dispatcher registration using slots: int (pre-slot-config engines)."""

import platform
from importlib.metadata import version
from sys import version_info
from typing import cast

from hatchet_sdk.clients.dispatcher.action_listener import ActionListener
from hatchet_sdk.config import ClientConfig
from hatchet_sdk.connection import new_conn
from hatchet_sdk.contracts.dispatcher_pb2 import (
    SDKS,
    RuntimeInfo,
    WorkerLabels,
    WorkerRegisterRequest,
    WorkerRegisterResponse,
)
from hatchet_sdk.contracts.dispatcher_pb2_grpc import DispatcherStub
from hatchet_sdk.deprecated.action_listener import LegacyGetActionListenerRequest
from hatchet_sdk.metadata import get_metadata

DEFAULT_REGISTER_TIMEOUT = 30


async def legacy_get_action_listener(
    config: ClientConfig,
    req: LegacyGetActionListenerRequest,
) -> ActionListener:
    """Register a worker using the legacy slots field (for pre-slot-config engines)."""
    aio_conn = new_conn(config, True)
    aio_client = DispatcherStub(aio_conn)

    # Override labels with the preset labels
    preset_labels = config.worker_preset_labels

    for key, value in preset_labels.items():
        req.labels[key] = WorkerLabels(str_value=str(value))

    response = cast(
        WorkerRegisterResponse,
        await aio_client.Register(  # type: ignore[misc]
            WorkerRegisterRequest(
                worker_name=req.worker_name,
                actions=req.actions,
                services=req.services,
                slots=req.slots,
                labels=req.labels,
                runtime_info=RuntimeInfo(
                    sdk_version=version("hatchet_sdk"),
                    language=SDKS.PYTHON,
                    language_version=f"{version_info.major}.{version_info.minor}.{version_info.micro}",
                    os=platform.system().lower(),
                ),
            ),
            timeout=DEFAULT_REGISTER_TIMEOUT,
            metadata=get_metadata(config.token),
        ),
    )

    return ActionListener(config, response.worker_id)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: dispatcher.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'dispatcher.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x64ispatcher.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"Z\n\x0cWorkerLabels\x12\x16\n\tstr_value\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x16\n\tint_value\x18\x02 \x01(\x05H\x01\x88\x01\x01\x42\x0c\n\n_str_valueB\x0c\n\n_int_value\"\xcc\x01\n\x0bRuntimeInfo\x12\x18\n\x0bsdk_version\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x1c\n\x08language\x18\x02 \x01(\x0e\x32\x05.SDKSH\x01\x88\x01\x01\x12\x1d\n\x10language_version\x18\x03 \x01(\tH\x02\x88\x01\x01\x12\x0f\n\x02os\x18\x04 \x01(\tH\x03\x88\x01\x01\x12\x12\n\x05\x65xtra\x18\x05 \x01(\tH\x04\x88\x01\x01\x42\x0e\n\x0c_sdk_versionB\x0b\n\t_languageB\x13\n\x11_language_versionB\x05\n\x03_osB\x08\n\x06_extra\"\xb1\x03\n\x15WorkerRegisterRequest\x12\x13\n\x0bworker_name\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x63tions\x18\x02 \x03(\t\x12\x10\n\x08services\x18\x03 \x03(\t\x12\x12\n\x05slots\x18\x04 \x01(\x05H\x00\x88\x01\x01\x12\x32\n\x06labels\x18\x05 \x03(\x0b\x32\".WorkerRegisterRequest.LabelsEntry\x12\x17\n\nwebhook_id\x18\x06 \x01(\tH\x01\x88\x01\x01\x12\'\n\x0cruntime_info\x18\x07 \x01(\x0b\x32\x0c.RuntimeInfoH\x02\x88\x01\x01\x12;\n\x0bslot_config\x18\t \x03(\x0b\x32&.WorkerRegisterRequest.SlotConfigEntry\x1a<\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1c\n\x05value\x18\x02 \x01(\x0b\x32\r.WorkerLabels:\x02\x38\x01\x1a\x31\n\x0fSlotConfigEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x42\x08\n\x06_slotsB\r\n\x0b_webhook_idB\x0f\n\r_runtime_info\"S\n\x16WorkerRegisterResponse\x12\x11\n\ttenant_id\x18\x01 \x01(\t\x12\x11\n\tworker_id\x18\x02 \x01(\t\x12\x13\n\x0bworker_name\x18\x03 \x01(\t\"\xa4\x01\n\x19UpsertWorkerLabelsRequest\x12\x11\n\tworker_id\x18\x01 \x01(\t\x12\x36\n\x06labels\x18\x02 \x03(\x0b\x32&.UpsertWorkerLabelsRequest.LabelsEntry\x1a<\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1c\n\x05value\x18\x02 \x01(\x0b\x32\r.WorkerLabels:\x02\x38\x01\"B\n\x1aUpsertWorkerLabelsResponse\x12\x11\n\ttenant_id\x18\x01 \x01(\t\x12\x11\n\tworker_id\x18\x02 \x01(\t\"\x98\x05\n\x0e\x41ssignedAction\x12\x11\n\ttenant_id\x18\x01 \x01(\t\x12\x17\n\x0fworkflow_run_id\x18\x02 \x01(\t\x12\x1c\n\x14get_group_key_run_id\x18\x03 \x01(\t\x12\x0e\n\x06job_id\x18\x04 \x01(\t\x12\x10\n\x08job_name\x18\x05 \x01(\t\x12\x12\n\njob_run_id\x18\x06 \x01(\t\x12\x0f\n\x07task_id\x18\x07 \x01(\t\x12\x1c\n\x14task_run_external_id\x18\x08 \x01(\t\x12\x11\n\taction_id\x18\t \x01(\t\x12 \n\x0b\x61\x63tion_type\x18\n \x01(\x0e\x32\x0b.ActionType\x12\x16\n\x0e\x61\x63tion_payload\x18\x0b \x01(\t\x12\x11\n\ttask_name\x18\x0c \x01(\t\x12\x13\n\x0bretry_count\x18\r \x01(\x05\x12 \n\x13\x61\x64\x64itional_metadata\x18\x0e \x01(\tH\x00\x88\x01\x01\x12!\n\x14\x63hild_workflow_index\x18\x0f \x01(\x05H\x01\x88\x01\x01\x12\x1f\n\x12\x63hild_workflow_key\x18\x10 \x01(\tH\x02\x88\x01\x01\x12#\n\x16parent_workflow_run_id\x18\x11 \x01(\tH\x03\x88\x01\x01\x12\x10\n\x08priority\x18\x12 \x01(\x05\x12\x18\n\x0bworkflow_id\x18\x13 \x01(\tH\x04\x88\x01\x01\x12 \n\x13workflow_version_id\x18\x14 \x01(\tH\x05\x88\x01\x01\x42\x16\n\x14_additional_metadataB\x17\n\x15_child_workflow_indexB\x15\n\x13_child_workflow_keyB\x19\n\x17_parent_workflow_run_idB\x0e\n\x0c_workflow_idB\x16\n\x14_workflow_version_id\"(\n\x13WorkerListenRequest\x12\x11\n\tworker_id\x18\x01 \x01(\t\"-\n\x18WorkerUnsubscribeRequest\x12\x11\n\tworker_id\x18\x01 \x01(\t\"A\n\x19WorkerUnsubscribeResponse\x12\x11\n\ttenant_id\x18\x01 \x01(\t\x12\x11\n\tworker_id\x18\x02 \x01(\t\"\xec\x01\n\x13GroupKeyActionEvent\x12\x11\n\tworker_id\x18\x01 \x01(\t\x12\x17\n\x0fworkflow_run_id\x18\x02 \x01(\t\x12\x1c\n\x14get_group_key_run_id\x18\x03 \x01(\t\x12\x11\n\taction_id\x18\x04 \x01(\t\x12\x33\n\x0f\x65vent_timestamp\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\nevent_type\x18\x06 \x01(\x0e\x32\x18.GroupKeyActionEventType\x12\x15\n\revent_payload\x18\x07 \x01(\t\"\xde\x02\n\x0fStepActionEvent\x12\x11\n\tworker_id\x18\x01 \x01(\t\x12\x0e\n\x06job_id\x18\x02 \x01(\t\x12\x12\n\njob_run_id\x18\x03 \x01(\t\x12\x0f\n\x07task_id\x18\x04 \x01(\t\x12\x1c\n\x14task_run_external_id\x18\x05 \x01(\t\x12\x11\n\taction_id\x18\x06 \x01(\t\x12\x33\n\x0f\x65vent_timestamp\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12(\n\nevent_type\x18\x08 \x01(\x0e\x32\x14.StepActionEventType\x12\x15\n\revent_payload\x18\t \x01(\t\x12\x18\n\x0bretry_count\x18\n \x01(\x05H\x00\x88\x01\x01\x12\x1d\n\x10should_not_retry\x18\x0b \x01(\x08H\x01\x88\x01\x01\x42\x0e\n\x0c_retry_countB\x13\n\x11_should_not_retry\";\n\x13\x41\x63tionEventResponse\x12\x11\n\ttenant_id\x18\x01 \x01(\t\x12\x11\n\tworker_id\x18\x02 \x01(\t\"\xcc\x01\n SubscribeToWorkflowEventsRequest\x12\x1c\n\x0fworkflow_run_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12 \n\x13\x61\x64\x64itional_meta_key\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\"\n\x15\x61\x64\x64itional_meta_value\x18\x03 \x01(\tH\x02\x88\x01\x01\x42\x12\n\x10_workflow_run_idB\x16\n\x14_additional_meta_keyB\x18\n\x16_additional_meta_value\"9\n\x1eSubscribeToWorkflowRunsRequest\x12\x17\n\x0fworkflow_run_id\x18\x01 \x01(\t\"\xe7\x02\n\rWorkflowEvent\x12\x17\n\x0fworkflow_run_id\x18\x01 \x01(\t\x12$\n\rresource_type\x18\x02 \x01(\x0e\x32\r.ResourceType\x12&\n\nevent_type\x18\x03 \x01(\x0e\x32\x12.ResourceEventType\x12\x13\n\x0bresource_id\x18\x04 \x01(\t\x12\x33\n\x0f\x65vent_timestamp\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x15\n\revent_payload\x18\x06 \x01(\t\x12\x0e\n\x06hangup\x18\x07 \x01(\x08\x12\x19\n\x0ctask_retries\x18\x08 \x01(\x05H\x00\x88\x01\x01\x12\x18\n\x0bretry_count\x18\t \x01(\x05H\x01\x88\x01\x01\x12\x18\n\x0b\x65vent_index\x18\n \x01(\x03H\x02\x88\x01\x01\x42\x0f\n\r_task_retriesB\x0e\n\x0c_retry_countB\x0e\n\x0c_event_index\"\xac\x01\n\x10WorkflowRunEvent\x12\x17\n\x0fworkflow_run_id\x18\x01 \x01(\t\x12)\n\nevent_type\x18\x02 \x01(\x0e\x32\x15.WorkflowRunEventType\x12\x33\n\x0f\x65vent_timestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1f\n\x07results\x18\x04 \x03(\x0b\x32\x0e.StepRunResult\"\x92\x01\n\rStepRunResult\x12\x1c\n\x14task_run_external_id\x18\x01 \x01(\t\x12\x11\n\ttask_name\x18\x02 \x01(\t\x12\x12\n\njob_run_id\x18\x03 \x01(\t\x12\x12\n\x05\x65rror\x18\x04 \x01(\tH\x00\x88\x01\x01\x12\x13\n\x06output\x18\x05 \x01(\tH\x01\x88\x01\x01\x42\x08\n\x06_errorB\t\n\x07_output\"c\n\rOverridesData\x12\x1c\n\x14task_run_external_id\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\x12\x17\n\x0f\x63\x61ller_filename\x18\x04 \x01(\t\"\x17\n\x15OverridesDataResponse\"W\n\x10HeartbeatRequest\x12\x11\n\tworker_id\x18\x01 \x01(\t\x12\x30\n\x0cheartbeat_at\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x13\n\x11HeartbeatResponse\"S\n\x15RefreshTimeoutRequest\x12\x1c\n\x14task_run_external_id\x18\x01 \x01(\t\x12\x1c\n\x14increment_timeout_by\x18\x02 \x01(\t\"H\n\x16RefreshTimeoutResponse\x12.\n\ntimeout_at\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"2\n\x12ReleaseSlotRequest\x12\x1c\n\x14task_run_external_id\x18\x01 \x01(\t\"\x15\n\x13ReleaseSlotResponse\"\x13\n\x11GetVersionRequest\"%\n\x12GetVersionResponse\x12\x0f\n\x07version\x18\x01 \x01(\t*A\n\x04SDKS\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x06\n\x02GO\x10\x01\x12\n\n\x06PYTHON\x10\x02\x12\x0e\n\nTYPESCRIPT\x10\x03\x12\x08\n\x04RUBY\x10\x04*N\n\nActionType\x12\x12\n\x0eSTART_STEP_RUN\x10\x00\x12\x13\n\x0f\x43\x41NCEL_STEP_RUN\x10\x01\x12\x17\n\x13START_GET_GROUP_KEY\x10\x02*\xa2\x01\n\x17GroupKeyActionEventType\x12 \n\x1cGROUP_KEY_EVENT_TYPE_UNKNOWN\x10\x00\x12 \n\x1cGROUP_KEY_EVENT_TYPE_STARTED\x10\x01\x12\"\n\x1eGROUP_KEY_EVENT_TYPE_COMPLETED\x10\x02\x12\x1f\n\x1bGROUP_KEY_EVENT_TYPE_FAILED\x10\x03*\xac\x01\n\x13StepActionEventType\x12\x1b\n\x17STEP_EVENT_TYPE_UNKNOWN\x10\x00\x12\x1b\n\x17STEP_EVENT_TYPE_STARTED\x10\x01\x12\x1d\n\x19STEP_EVENT_TYPE_COMPLETED\x10\x02\x12\x1a\n\x16STEP_EVENT_TYPE_FAILED\x10\x03\x12 \n\x1cSTEP_EVENT_TYPE_ACKNOWLEDGED\x10\x04*e\n\x0cResourceType\x12\x19\n\x15RESOURCE_TYPE_UNKNOWN\x10\x00\x12\x1a\n\x16RESOURCE_TYPE_STEP_RUN\x10\x01\x12\x1e\n\x1aRESOURCE_TYPE_WORKFLOW_RUN\x10\x02*\xfe\x01\n\x11ResourceEventType\x12\x1f\n\x1bRESOURCE_EVENT_TYPE_UNKNOWN\x10\x00\x12\x1f\n\x1bRESOURCE_EVENT_TYPE_STARTED\x10\x01\x12!\n\x1dRESOURCE_EVENT_TYPE_COMPLETED\x10\x02\x12\x1e\n\x1aRESOURCE_EVENT_TYPE_FAILED\x10\x03\x12!\n\x1dRESOURCE_EVENT_TYPE_CANCELLED\x10\x04\x12!\n\x1dRESOURCE_EVENT_TYPE_TIMED_OUT\x10\x05\x12\x1e\n\x1aRESOURCE_EVENT_TYPE_STREAM\x10\x06*<\n\x14WorkflowRunEventType\x12$\n WORKFLOW_RUN_EVENT_TYPE_FINISHED\x10\x00\x32\xb1\x07\n\nDispatcher\x12=\n\x08Register\x12\x16.WorkerRegisterRequest\x1a\x17.WorkerRegisterResponse\"\x00\x12\x33\n\x06Listen\x12\x14.WorkerListenRequest\x1a\x0f.AssignedAction\"\x00\x30\x01\x12\x35\n\x08ListenV2\x12\x14.WorkerListenRequest\x1a\x0f.AssignedAction\"\x00\x30\x01\x12\x34\n\tHeartbeat\x12\x11.HeartbeatRequest\x1a\x12.HeartbeatResponse\"\x00\x12R\n\x19SubscribeToWorkflowEvents\x12!.SubscribeToWorkflowEventsRequest\x1a\x0e.WorkflowEvent\"\x00\x30\x01\x12S\n\x17SubscribeToWorkflowRuns\x12\x1f.SubscribeToWorkflowRunsRequest\x1a\x11.WorkflowRunEvent\"\x00(\x01\x30\x01\x12?\n\x13SendStepActionEvent\x12\x10.StepActionEvent\x1a\x14.ActionEventResponse\"\x00\x12G\n\x17SendGroupKeyActionEvent\x12\x14.GroupKeyActionEvent\x1a\x14.ActionEventResponse\"\x00\x12<\n\x10PutOverridesData\x12\x0e.OverridesData\x1a\x16.OverridesDataResponse\"\x00\x12\x46\n\x0bUnsubscribe\x12\x19.WorkerUnsubscribeRequest\x1a\x1a.WorkerUnsubscribeResponse\"\x00\x12\x43\n\x0eRefreshTimeout\x12\x16.RefreshTimeoutRequest\x1a\x17.RefreshTimeoutResponse\"\x00\x12:\n\x0bReleaseSlot\x12\x13.ReleaseSlotRequest\x1a\x14.ReleaseSlotResponse\"\x00\x12O\n\x12UpsertWorkerLabels\x12\x1a.UpsertWorkerLabelsRequest\x1a\x1b.UpsertWorkerLabelsResponse\"\x00\x12\x37\n\nGetVersion\x12\x12.GetVersionRequest\x1a\x13.GetVersionResponse\"\x00\x42GZEgithub.com/hatchet-dev/hatchet/internal/services/dispatcher/contractsb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'dispatcher_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'ZEgithub.com/hatchet-dev/hatchet/internal/services/dispatcher/contracts'
  _globals['_WORKERREGISTERREQUEST_LABELSENTRY']._loaded_options = None
  _globals['_WORKERREGISTERREQUEST_LABELSENTRY']._serialized_options = b'8\001'
  _globals['_WORKERREGISTERREQUEST_SLOTCONFIGENTRY']._loaded_options = None
  _globals['_WORKERREGISTERREQUEST_SLOTCONFIGENTRY']._serialized_options = b'8\001'
  _globals['_UPSERTWORKERLABELSREQUEST_LABELSENTRY']._loaded_options = None
  _globals['_UPSERTWORKERLABELSREQUEST_LABELSENTRY']._serialized_options = b'8\001'
  _globals['_SDKS']._serialized_start=4066
  _globals['_SDKS']._serialized_end=4131
  _globals['_ACTIONTYPE']._serialized_start=4133
  _globals['_ACTIONTYPE']._serialized_end=4211
  _globals['_GROUPKEYACTIONEVENTTYPE']._serialized_start=4214
  _globals['_GROUPKEYACTIONEVENTTYPE']._serialized_end=4376
  _globals['_STEPACTIONEVENTTYPE']._serialized_start=4379
  _globals['_STEPACTIONEVENTTYPE']._serialized_end=4551
  _globals['_RESOURCETYPE']._serialized_start=4553
  _globals['_RESOURCETYPE']._serialized_end=4654
  _globals['_RESOURCEEVENTTYPE']._serialized_start=4657
  _globals['_RESOURCEEVENTTYPE']._serialized_end=4911
  _globals['_WORKFLOWRUNEVENTTYPE']._serialized_start=4913
  _globals['_WORKFLOWRUNEVENTTYPE']._serialized_end=4973
  _globals['_WORKERLABELS']._serialized_start=53
  _globals['_WORKERLABELS']._serialized_end=143
  _globals['_RUNTIMEINFO']._serialized_start=146
  _globals['_RUNTIMEINFO']._serialized_end=350
  _globals['_WORKERREGISTERREQUEST']._serialized_start=353
  _globals['_WORKERREGISTERREQUEST']._serialized_end=786
  _globals['_WORKERREGISTERREQUEST_LABELSENTRY']._serialized_start=633
  _globals['_WORKERREGISTERREQUEST_LABELSENTRY']._serialized_end=693
  _globals['_WORKERREGISTERREQUEST_SLOTCONFIGENTRY']._serialized_start=695
  _globals['_WORKERREGISTERREQUEST_SLOTCONFIGENTRY']._serialized_end=744
  _globals['_WORKERREGISTERRESPONSE']._serialized_start=788
  _globals['_WORKERREGISTERRESPONSE']._serialized_end=871
  _globals['_UPSERTWORKERLABELSREQUEST']._serialized_start=874
  _globals['_UPSERTWORKERLABELSREQUEST']._serialized_end=1038
  _globals['_UPSERTWORKERLABELSREQUEST_LABELSENTRY']._serialized_start=633
  _globals['_UPSERTWORKERLABELSREQUEST_LABELSENTRY']._serialized_end=693
  _globals['_UPSERTWORKERLABELSRESPONSE']._serialized_start=1040
  _globals['_UPSERTWORKERLABELSRESPONSE']._serialized_end=1106
  _globals['_ASSIGNEDACTION']._serialized_start=1109
  _globals['_ASSIGNEDACTION']._serialized_end=1773
  _globals['_WORKERLISTENREQUEST']._serialized_start=1775
  _globals['_WORKERLISTENREQUEST']._serialized_end=1815
  _globals['_WORKERUNSUBSCRIBEREQUEST']._serialized_start=1817
  _globals['_WORKERUNSUBSCRIBEREQUEST']._serialized_end=1862
  _globals['_WORKERUNSUBSCRIBERESPONSE']._serialized_start=1864
  _globals['_WORKERUNSUBSCRIBERESPONSE']._serialized_end=1929
  _globals['_GROUPKEYACTIONEVENT']._serialized_start=1932
  _globals['_GROUPKEYACTIONEVENT']._serialized_end=2168
  _globals['_STEPACTIONEVENT']._serialized_start=2171
  _globals['_STEPACTIONEVENT']._serialized_end=2521
  _globals['_ACTIONEVENTRESPONSE']._serialized_start=2523
  _globals['_ACTIONEVENTRESPONSE']._serialized_end=2582
  _globals['_SUBSCRIBETOWORKFLOWEVENTSREQUEST']._serialized_start=2585
  _globals['_SUBSCRIBETOWORKFLOWEVENTSREQUEST']._serialized_end=2789
  _globals['_SUBSCRIBETOWORKFLOWRUNSREQUEST']._serialized_start=2791
  _globals['_SUBSCRIBETOWORKFLOWRUNSREQUEST']._serialized_end=2848
  _globals['_WORKFLOWEVENT']._serialized_start=2851
  _globals['_WORKFLOWEVENT']._serialized_end=3210
  _globals['_WORKFLOWRUNEVENT']._serialized_start=3213
  _globals['_WORKFLOWRUNEVENT']._serialized_end=3385
  _globals['_STEPRUNRESULT']._serialized_start=3388
  _globals['_STEPRUNRESULT']._serialized_end=3534
  _globals['_OVERRIDESDATA']._serialized_start=3536
  _globals['_OVERRIDESDATA']._serialized_end=3635
  _globals['_OVERRIDESDATARESPONSE']._serialized_start=3637
  _globals['_OVERRIDESDATARESPONSE']._serialized_end=3660
  _globals['_HEARTBEATREQUEST']._serialized_start=3662
  _globals['_HEARTBEATREQUEST']._serialized_end=3749
  _globals['_HEARTBEATRESPONSE']._serialized_start=3751
  _globals['_HEARTBEATRESPONSE']._serialized_end=3770
  _globals['_REFRESHTIMEOUTREQUEST']._serialized_start=3772
  _globals['_REFRESHTIMEOUTREQUEST']._serialized_end=3855
  _globals['_REFRESHTIMEOUTRESPONSE']._serialized_start=3857
  _globals['_REFRESHTIMEOUTRESPONSE']._serialized_end=3929
  _globals['_RELEASESLOTREQUEST']._serialized_start=3931
  _globals['_RELEASESLOTREQUEST']._serialized_end=3981
  _globals['_RELEASESLOTRESPONSE']._serialized_start=3983
  _globals['_RELEASESLOTRESPONSE']._serialized_end=4004
  _globals['_GETVERSIONREQUEST']._serialized_start=4006
  _globals['_GETVERSIONREQUEST']._serialized_end=4025
  _globals['_GETVERSIONRESPONSE']._serialized_start=4027
  _globals['_GETVERSIONRESPONSE']._serialized_end=4064
  _globals['_DISPATCHER']._serialized_start=4976
  _globals['_DISPATCHER']._serialized_end=5921
# @@protoc_insertion_point(module_scope)



"""Legacy GetActionListenerRequest using slots: int (pre-slot-config engines)."""

from pydantic import BaseModel, ConfigDict, Field, model_validator

from hatchet_sdk.contracts.dispatcher_pb2 import WorkerLabels


class LegacyGetActionListenerRequest(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    worker_name: str
    services: list[str]
    actions: list[str]
    slots: int
    raw_labels: dict[str, str | int] = Field(default_factory=dict)

    labels: dict[str, WorkerLabels] = Field(default_factory=dict)

    @model_validator(mode="after")
    def validate_labels(self) -> "LegacyGetActionListenerRequest":
        self.labels = {}

        for key, value in self.raw_labels.items():
            if isinstance(value, int):
                self.labels[key] = WorkerLabels(int_value=value)
            else:
                self.labels[key] = WorkerLabels(str_value=str(value))

        return self



"""Legacy action listener process using slots: int (pre-slot-config engines)."""

import asyncio
import contextlib
import logging
import signal
import time
from datetime import timedelta
from multiprocessing import Queue

import grpc
from aiohttp import web
from aiohttp.web_request import Request
from aiohttp.web_response import Response
from grpc.aio import UnaryUnaryCall
from prometheus_client import Gauge, generate_latest

from hatchet_sdk.client import Client
from hatchet_sdk.clients.dispatcher.action_listener import ActionListener
from hatchet_sdk.clients.rest.models.update_worker_request import UpdateWorkerRequest
from hatchet_sdk.config import ClientConfig
from hatchet_sdk.contracts.dispatcher_pb2 import (
    STEP_EVENT_TYPE_STARTED,
    ActionEventResponse,
    StepActionEvent,
)
from hatchet_sdk.deprecated.action_listener import LegacyGetActionListenerRequest
from hatchet_sdk.deprecated.dispatcher import legacy_get_action_listener
from hatchet_sdk.logger import logger
from hatchet_sdk.runnables.action import Action, ActionType
from hatchet_sdk.runnables.contextvars import (
    ctx_action_key,
    ctx_step_run_id,
    ctx_task_retry_count,
    ctx_worker_id,
    ctx_workflow_run_id,
)
from hatchet_sdk.utils.backoff import exp_backoff_sleep
from hatchet_sdk.utils.typing import STOP_LOOP, STOP_LOOP_TYPE
from hatchet_sdk.worker.action_listener_process import (
    BLOCKED_THREAD_WARNING,
    STARTING_UNHEALTHY_AFTER_SECONDS,
    ActionEvent,
    HealthStatus,
)

ACTION_EVENT_RETRY_COUNT = 5


class LegacyWorkerActionListenerProcess:
    """Worker action listener process that uses the legacy slots: int registration."""

    def __init__(
        self,
        name: str,
        actions: list[str],
        slots: int,
        config: ClientConfig,
        action_queue: "Queue[Action]",
        event_queue: "Queue[ActionEvent | STOP_LOOP_TYPE]",
        handle_kill: bool,
        debug: bool,
        labels: dict[str, str | int],
    ) -> None:
        self.name = name
        self.actions = actions
        self.slots = slots
        self.config = config
        self.action_queue = action_queue
        self.event_queue = event_queue
        self.debug = debug
        self.labels = labels
        self.handle_kill = handle_kill

        self._health_runner: web.AppRunner | None = None
        self._listener_health_gauge: Gauge | None = None
        self._event_loop_lag_gauge: Gauge | None = None
        self._event_loop_monitor_task: asyncio.Task[None] | None = None
        self._event_loop_last_lag_seconds: float = 0.0
        self._event_loop_blocked_since: float | None = None
        self._waiting_steps_blocked_since: float | None = None
        self._starting_since: float = time.time()

        self.listener: ActionListener | None = None
        self.killing = False
        self.action_loop_task: asyncio.Task[None] | None = None
        self.event_send_loop_task: asyncio.Task[None] | None = None
        self.running_step_runs: dict[str, float] = {}
        self.step_action_events: set[
            asyncio.Task[UnaryUnaryCall[StepActionEvent, ActionEventResponse] | None]
        ] = set()

        if self.debug:
            logger.setLevel(logging.DEBUG)

        self.client = Client(config=self.config, debug=self.debug)

        loop = asyncio.get_event_loop()
        loop.add_signal_handler(
            signal.SIGINT, lambda: asyncio.create_task(self.pause_task_assignment())
        )
        loop.add_signal_handler(
            signal.SIGTERM, lambda: asyncio.create_task(self.pause_task_assignment())
        )
        loop.add_signal_handler(
            signal.SIGQUIT, lambda: asyncio.create_task(self.exit_gracefully())
        )

        if self.config.healthcheck.enabled:
            self._listener_health_gauge = Gauge(
                "hatchet_worker_listener_health",
                "Listener health (1 healthy, 0 unhealthy)",
            )
            self._event_loop_lag_gauge = Gauge(
                "hatchet_worker_event_loop_lag_seconds",
                "Event loop lag in seconds (listener process)",
            )

    async def _monitor_event_loop(self) -> None:
        interval = 0.5
        while not self.killing:
            start = time.time()
            await asyncio.sleep(interval)
            elapsed = time.time() - start
            lag = max(0.0, elapsed - interval)
            if (
                timedelta(seconds=lag)
                >= self.config.healthcheck.event_loop_block_threshold_seconds
            ):
                if self._event_loop_blocked_since is None:
                    self._event_loop_blocked_since = start + interval
                self._event_loop_last_lag_seconds = max(
                    lag, time.time() - self._event_loop_blocked_since
                )
            else:
                self._event_loop_last_lag_seconds = lag

            if (
                timedelta(seconds=lag)
                < self.config.healthcheck.event_loop_block_threshold_seconds
            ):
                self._event_loop_blocked_since = None

    def _starting_timed_out(self) -> bool:
        return (time.time() - self._starting_since) > STARTING_UNHEALTHY_AFTER_SECONDS

    def _compute_health(self) -> HealthStatus:
        if self.killing:
            return HealthStatus.UNHEALTHY

        if (
            self._event_loop_blocked_since is not None
            and timedelta(seconds=(time.time() - self._event_loop_blocked_since))
            > self.config.healthcheck.event_loop_block_threshold_seconds
        ):
            return HealthStatus.UNHEALTHY

        if (
            self._waiting_steps_blocked_since is not None
            and timedelta(seconds=(time.time() - self._waiting_steps_blocked_since))
            > self.config.healthcheck.event_loop_block_threshold_seconds
        ):
            return HealthStatus.UNHEALTHY

        if self.listener is None:
            if self._starting_timed_out():
                return HealthStatus.UNHEALTHY
            return HealthStatus.STARTING

        listener = self.listener

        last_attempt = listener.last_connection_attempt or 0.0
        if last_attempt <= 0:
            if self._starting_timed_out():
                return HealthStatus.UNHEALTHY
            return HealthStatus.STARTING

        if listener.listen_strategy == "v2":
            now = time.time()
            time_last_hb = listener.time_last_hb_succeeded or 0.0
            has_hb_success = 0.0 < time_last_hb <= now
            ok = bool(
                listener.heartbeat_task is not None
                and listener.last_heartbeat_succeeded
                and has_hb_success
            )
        else:
            ok = bool(listener.retries == 0)

        return HealthStatus.HEALTHY if ok else HealthStatus.UNHEALTHY

    async def _health_handler(self, request: Request) -> Response:
        status = self._compute_health()
        ok = status == HealthStatus.HEALTHY
        response = {"status": status.value}
        return web.json_response(response, status=200 if ok else 503)

    async def _metrics_handler(self, request: Request) -> Response:
        status = self._compute_health()
        ok = status == HealthStatus.HEALTHY

        if self._listener_health_gauge is not None:
            self._listener_health_gauge.set(1 if ok else 0)

        if self._event_loop_lag_gauge is not None:
            self._event_loop_lag_gauge.set(self._event_loop_last_lag_seconds)

        return web.Response(body=generate_latest(), content_type="text/plain")

    async def start_health_server(self) -> None:
        if not self.config.healthcheck.enabled:
            return

        if self._health_runner is not None:
            return

        app = web.Application()
        app.add_routes(
            [
                web.get("/health", self._health_handler),
                web.get("/metrics", self._metrics_handler),
            ]
        )

        runner = web.AppRunner(app)

        try:
            await runner.setup()
            await web.TCPSite(
                runner,
                host=self.config.healthcheck.bind_address,
                port=self.config.healthcheck.port,
            ).start()
        except Exception:
            logger.exception("failed to start healthcheck server (listener process)")
            return

        self._health_runner = runner
        logger.info(
            f"healthcheck server (listener process) running on {self.config.healthcheck.bind_address}:{self.config.healthcheck.port}"
        )

        if self._event_loop_monitor_task is None:
            self._event_loop_monitor_task = asyncio.create_task(
                self._monitor_event_loop()
            )

    async def stop_health_server(self) -> None:
        if self._event_loop_monitor_task is not None:
            task = self._event_loop_monitor_task
            self._event_loop_monitor_task = None
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task

        if self._health_runner is None:
            return

        try:
            await self._health_runner.cleanup()
        except Exception:
            logger.exception("failed to stop healthcheck server (listener process)")
        finally:
            self._health_runner = None

    async def pause_task_assignment(self) -> None:
        if self.listener is None:
            raise ValueError("listener not started")

        await self.client.workers.aio_update(
            worker_id=self.listener.worker_id,
            opts=UpdateWorkerRequest(isPaused=True),
        )

    async def start(self, retry_attempt: int = 0) -> None:
        if retry_attempt > 5:
            logger.error("could not start action listener")
            return

        logger.debug(f"starting action listener: {self.name}")

        try:
            from hatchet_sdk.clients.dispatcher.dispatcher import DispatcherClient

            self.dispatcher_client = DispatcherClient(self.config)

            self.listener = await legacy_get_action_listener(
                self.config,
                LegacyGetActionListenerRequest(
                    worker_name=self.name,
                    services=["default"],
                    actions=self.actions,
                    slots=self.slots,
                    raw_labels=self.labels,
                ),
            )

            logger.debug(f"acquired action listener: {self.listener.worker_id}")
        except grpc.RpcError:
            logger.exception("could not start action listener")
            return

        self.action_loop_task = asyncio.create_task(self.start_action_loop())
        self.event_send_loop_task = asyncio.create_task(self.start_event_send_loop())
        self.blocked_main_loop = asyncio.create_task(self.start_blocked_main_loop())

    async def _get_event(self) -> ActionEvent | STOP_LOOP_TYPE:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self.event_queue.get)

    async def start_event_send_loop(self) -> None:
        while True:
            event = await self._get_event()
            if event == STOP_LOOP:
                logger.debug("stopping event send loop...")
                break

            logger.debug(f"tx: event: {event.action.action_id}/{event.type}")
            t = asyncio.create_task(self.send_event(event))
            self.step_action_events.add(t)
            t.add_done_callback(lambda t: self.step_action_events.discard(t))

    async def start_blocked_main_loop(self) -> None:
        threshold = 1
        while not self.killing:
            count = 0
            for start_time in self.running_step_runs.values():
                diff = self.now() - start_time
                if diff > threshold:
                    count += 1

            if count > 0:
                if self._waiting_steps_blocked_since is None:
                    self._waiting_steps_blocked_since = time.time()
                blocked_for = time.time() - self._waiting_steps_blocked_since
                logger.warning(
                    f"{BLOCKED_THREAD_WARNING} Waiting Steps {count} blocked_for={blocked_for:.1f}s"
                )
            else:
                self._waiting_steps_blocked_since = None
            await asyncio.sleep(1)

    async def send_event(self, event: ActionEvent, retry_attempt: int = 1) -> None:
        try:
            match event.action.action_type:
                case ActionType.START_STEP_RUN:
                    if event.type == STEP_EVENT_TYPE_STARTED:
                        if event.action.step_run_id in self.running_step_runs:
                            diff = (
                                self.now()
                                - self.running_step_runs[event.action.step_run_id]
                            )
                            if diff > 0.1:
                                logger.warning(
                                    f"{BLOCKED_THREAD_WARNING} time to start: {diff}s"
                                )
                            else:
                                logger.debug(f"start time: {diff}")
                            del self.running_step_runs[event.action.step_run_id]
                        else:
                            self.running_step_runs[event.action.step_run_id] = (
                                self.now()
                            )

                    send_started_event_task = asyncio.create_task(
                        self.dispatcher_client.send_step_action_event(
                            event.action,
                            event.type,
                            event.payload,
                            event.should_not_retry,
                        )
                    )

                    self.step_action_events.add(send_started_event_task)
                    send_started_event_task.add_done_callback(
                        lambda t: self.step_action_events.discard(t)
                    )
                case ActionType.CANCEL_STEP_RUN:
                    logger.debug("unimplemented event send")
                case _:
                    logger.error("unknown action type for event send")
        except Exception:
            logger.exception(
                f"could not send action event ({retry_attempt}/{ACTION_EVENT_RETRY_COUNT})"
            )
            if retry_attempt <= ACTION_EVENT_RETRY_COUNT:
                await exp_backoff_sleep(retry_attempt, 1)
                await self.send_event(event, retry_attempt + 1)

    def now(self) -> float:
        return time.time()

    async def start_action_loop(self) -> None:
        if self.listener is None:
            raise ValueError("listener not started")

        try:
            async for action in self.listener:
                if action is None:
                    break

                ctx_step_run_id.set(action.step_run_id)
                ctx_workflow_run_id.set(action.workflow_run_id)
                ctx_worker_id.set(action.worker_id)
                ctx_action_key.set(action.key)
                ctx_task_retry_count.set(action.retry_count)

                match action.action_type:
                    case ActionType.START_STEP_RUN:
                        self.event_queue.put(
                            ActionEvent(
                                action=action,
                                type=STEP_EVENT_TYPE_STARTED,
                                payload=None,
                                should_not_retry=False,
                            )
                        )
                        logger.info(
                            f"rx: start step run: {action.step_run_id}/{action.action_id}"
                        )

                        if action.step_run_id in self.running_step_runs:
                            logger.warning(
                                f"step run already running: {action.step_run_id}"
                            )

                    case ActionType.CANCEL_STEP_RUN:
                        logger.info(f"rx: cancel step run: {action.step_run_id}")
                    case _:
                        logger.error(
                            f"rx: unknown action type ({action.action_type}): {action.action_type}"
                        )
                try:
                    self.action_queue.put(action)
                except Exception:
                    logger.exception("error putting action")

        except Exception:
            logger.exception("error in action loop")
        finally:
            logger.info("action loop closed")
            if not self.killing:
                await self.exit_gracefully()

    async def cleanup(self) -> None:
        self.killing = True

        await self.stop_health_server()

        if self.listener is not None:
            self.listener.cleanup()

        self.event_queue.put(STOP_LOOP)

    async def exit_gracefully(self) -> None:
        if self.listener:
            self.listener.stop_signal = True

        if self.listener is not None:
            try:
                await self.pause_task_assignment()
            except Exception:
                logger.debug("failed to pause task assignment during graceful exit")

        if self.killing:
            return

        logger.debug("closing action listener...")

        await self.cleanup()

        while not self.event_queue.empty():
            pass

        logger.info("action listener closed")

    def exit_forcefully(self) -> None:
        asyncio.run(self.cleanup())
        logger.debug("forcefully closing listener...")


def legacy_worker_action_listener_process(
    name: str,
    actions: list[str],
    slots: int,
    config: ClientConfig,
    action_queue: "Queue[Action]",
    event_queue: "Queue[ActionEvent | STOP_LOOP_TYPE]",
    handle_kill: bool,
    debug: bool,
    labels: dict[str, str | int],
) -> None:
    async def run() -> None:
        process = LegacyWorkerActionListenerProcess(
            name=name,
            actions=actions,
            slots=slots,
            config=config,
            action_queue=action_queue,
            event_queue=event_queue,
            handle_kill=handle_kill,
            debug=debug,
            labels=labels,
        )
        await process.start_health_server()
        await process.start()
        while not process.killing:  # noqa: ASYNC110
            await asyncio.sleep(0.1)

    asyncio.run(run())



# coding: utf-8

"""
Hatchet API

The Hatchet API

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from hatchet_sdk.clients.rest.models.concurrency_limit_strategy import (
    ConcurrencyLimitStrategy,
)
from hatchet_sdk.clients.rest.models.concurrency_scope import ConcurrencyScope
from typing import Optional, Set
from typing_extensions import Self


class ConcurrencySetting(BaseModel):
    """
    ConcurrencySetting
    """  # noqa: E501

    max_runs: StrictInt = Field(
        description="The maximum number of concurrent workflow runs.", alias="maxRuns"
    )
    limit_strategy: ConcurrencyLimitStrategy = Field(
        description="The strategy to use when the concurrency limit is reached.",
        alias="limitStrategy",
    )
    expression: StrictStr = Field(
        description="The concurrency expression, used to generate a key from task inputs, metadata, etc."
    )
    step_readable_id: Optional[StrictStr] = Field(
        default=None,
        description="The readable id of the step to which this concurrency setting applies.",
        alias="stepReadableId",
    )
    scope: ConcurrencyScope = Field(description="The scope of the concurrency setting.")
    __properties: ClassVar[List[str]] = [
        "maxRuns",
        "limitStrategy",
        "expression",
        "stepReadableId",
        "scope",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of ConcurrencySetting from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of ConcurrencySetting from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "maxRuns": obj.get("maxRuns"),
                "limitStrategy": obj.get("limitStrategy"),
                "expression": obj.get("expression"),
                "stepReadableId": obj.get("stepReadableId"),
                "scope": obj.get("scope"),
            }
        )
        return _obj



import datetime

from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Event(_message.Message):
    __slots__ = ("tenant_id", "event_id", "key", "payload", "event_timestamp", "additional_metadata", "scope")
    TENANT_ID_FIELD_NUMBER: _ClassVar[int]
    EVENT_ID_FIELD_NUMBER: _ClassVar[int]
    KEY_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    EVENT_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    ADDITIONAL_METADATA_FIELD_NUMBER: _ClassVar[int]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    tenant_id: str
    event_id: str
    key: str
    payload: str
    event_timestamp: _timestamp_pb2.Timestamp
    additional_metadata: str
    scope: str
    def __init__(self, tenant_id: _Optional[str] = ..., event_id: _Optional[str] = ..., key: _Optional[str] = ..., payload: _Optional[str] = ..., event_timestamp: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., additional_metadata: _Optional[str] = ..., scope: _Optional[str] = ...) -> None: ...

class Events(_message.Message):
    __slots__ = ("events",)
    EVENTS_FIELD_NUMBER: _ClassVar[int]
    events: _containers.RepeatedCompositeFieldContainer[Event]
    def __init__(self, events: _Optional[_Iterable[_Union[Event, _Mapping]]] = ...) -> None: ...

class PutLogRequest(_message.Message):
    __slots__ = ("task_run_external_id", "created_at", "message", "level", "metadata", "task_retry_count")
    TASK_RUN_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    LEVEL_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    TASK_RETRY_COUNT_FIELD_NUMBER: _ClassVar[int]
    task_run_external_id: str
    created_at: _timestamp_pb2.Timestamp
    message: str
    level: str
    metadata: str
    task_retry_count: int
    def __init__(self, task_run_external_id: _Optional[str] = ..., created_at: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., message: _Optional[str] = ..., level: _Optional[str] = ..., metadata: _Optional[str] = ..., task_retry_count: _Optional[int] = ...) -> None: ...

class PutLogResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class PutStreamEventRequest(_message.Message):
    __slots__ = ("task_run_external_id", "created_at", "message", "metadata", "event_index")
    TASK_RUN_EXTERNAL_ID_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    EVENT_INDEX_FIELD_NUMBER: _ClassVar[int]
    task_run_external_id: str
    created_at: _timestamp_pb2.Timestamp
    message: bytes
    metadata: str
    event_index: int
    def __init__(self, task_run_external_id: _Optional[str] = ..., created_at: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., message: _Optional[bytes] = ..., metadata: _Optional[str] = ..., event_index: _Optional[int] = ...) -> None: ...

class PutStreamEventResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class BulkPushEventRequest(_message.Message):
    __slots__ = ("events",)
    EVENTS_FIELD_NUMBER: _ClassVar[int]
    events: _containers.RepeatedCompositeFieldContainer[PushEventRequest]
    def __init__(self, events: _Optional[_Iterable[_Union[PushEventRequest, _Mapping]]] = ...) -> None: ...

class PushEventRequest(_message.Message):
    __slots__ = ("key", "payload", "event_timestamp", "additional_metadata", "priority", "scope")
    KEY_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    EVENT_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    ADDITIONAL_METADATA_FIELD_NUMBER: _ClassVar[int]
    PRIORITY_FIELD_NUMBER: _ClassVar[int]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    key: str
    payload: str
    event_timestamp: _timestamp_pb2.Timestamp
    additional_metadata: str
    priority: int
    scope: str
    def __init__(self, key: _Optional[str] = ..., payload: _Optional[str] = ..., event_timestamp: _Optional[_Union[datetime.datetime, _timestamp_pb2.Timestamp, _Mapping]] = ..., additional_metadata: _Optional[str] = ..., priority: _Optional[int] = ..., scope: _Optional[str] = ...) -> None: ...

class ReplayEventRequest(_message.Message):
    __slots__ = ("event_id",)
    EVENT_ID_FIELD_NUMBER: _ClassVar[int]
    event_id: str
    def __init__(self, event_id: _Optional[str] = ...) -> None: ...
